// SPDX-FileCopyrightText: 2018-2023 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef MARKERSWIDGET_H
#define MARKERSWIDGET_H

// Local includes
#include "shared/Markers.h"

// Qt includes
#include <QWidget>
#include <QHash>
#include <QVector>

// C++ includes
#include <utility>

// Local classes
class SharedObjects;
class Database;
class Settings;
class TournamentSettings;
class Server;
class Client;
class ChecksumHelper;
class MinimumViewportWidget;
class MinimumViewportScroll;
class MarkersListModel;
class MarkersListView;

// Qt classes
class QComboBox;
class QLineEdit;
class QGroupBox;
class QCheckBox;
class QLabel;

class MarkersWidget : public QWidget
{
    Q_OBJECT

public:
    enum MarkerData {
        Id = Qt::UserRole,
        Sorting
    };

    explicit MarkersWidget(QWidget *parent, SharedObjects *sharedObjects);
    void updateTournamentMode();
    void reload();
    void reset();
    void updateCounters(const QHash<int, int> &counter);
    int newRegistrationMarker(const QString &name) const;
    int drawnMarker() const;
    std::pair<int, QString> markerAfterEdit(const QString &oldName, const QString &name) const;
    void addChecksumData(ChecksumHelper &checksumHelper) const;
    void serverStarted(Server *server);
    void serverStopped();
    void clientConnected(Client *client);
    void clientDisconnected();
    void processNetworkChange();
    void setProcessingRequest(bool processing);
    void updateTournamentState();
    int singleMarker() const;
    int assignedMarker() const;
    void resetSingleConditionString();
    QString singleConditionString() const;
    void setCheckableMarker(Markers::Type type, int markerId);
    void setSinglesManagement(const Markers::SinglesManagement &data);
    void warnAboutSpecialMarkers();
    const QVector<Markers::Marker> &markers() const;
    const Markers::Marker &markerData(int id) const;
    const MarkersListModel *markersModel() const;

Q_SIGNALS:
    void markersChanged();
    void statusUpdate(const QString &text);
    void visibilityChanged();
    void updatePlayersList();

private Q_SLOTS:
    void addMarker();
    void saveNewMarker(const Markers::Marker &marker);
    void moveMarker(int direction);
    void editMarker();
    void saveEdit(const Markers::Marker &marker);
    void deleteMarker();
    void toggleAbsentMarker(bool state);
    void absentMarkerChanged();
    void singlesManagementChanged();
    void toggleDrawnMarker(bool state);
    void drawnMarkerChanged();

private: // Functions
    bool checkSelectedMarkerChanged(const QString &name, const QString &action);
    int markersComboBoxSelection(QComboBox *comboBox) const;
    void setMarkersComboBoxSelection(QComboBox *comboBox, const int markerId);

private: // Variables
    Database *m_db;
    Settings *m_settings;
    TournamentSettings *m_tournamentSettings;

    MarkersListModel *m_markersModel;
    MarkersListView *m_markersView;

    QHash<Markers::Type, QComboBox *> m_markersComboBoxes;
    QHash<Markers::Type, QCheckBox *> m_enableMarkerCheckBoxes;
    QGroupBox *m_automaticMarkersBox;
    QLabel *m_defaultMarkerLabel;
    QGroupBox *m_absentMarkerBox;
    QGroupBox *m_singlesManagementBox;
    QCheckBox *m_markSingles;
    QComboBox *m_singleCondition;
    QLineEdit *m_singleConditionString;
    MinimumViewportWidget *m_markerSettingsWidget;
    MinimumViewportScroll *m_settingsScroll;

    bool m_specialMarkersWarnigAlreadyShown = false;

    Server *m_server = nullptr;
    Client *m_client = nullptr;
    bool m_isServer = false;
    bool m_isClient = false;
};

#endif // MARKERSWIDGET_H
