// SPDX-FileCopyrightText: 2023 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "RegistrationListDelegate.h"
#include "PlayersEditor.h"

#include "shared/Logging.h"
#include "shared/RegistrationColumns.h"

// Qt includes
#include <QDebug>
#include <QPainter>
#include <QApplication>
#include <QTimer>

RegistrationListDelegate::RegistrationListDelegate(QObject *parent) : TableDelegate(parent)
{
    const auto darkMode = qApp->property("DARK_MODE").toBool();

    const QPalette palette;

    if (! darkMode) {
        m_highlightBrush = QBrush(palette.color(QPalette::Highlight).lighter(180));
    } else {
        m_highlightBrush = QBrush(palette.color(QPalette::Highlight).darker(330));
    }

    m_selectedPen = QPen(palette.color(QPalette::Highlight), 2);
}

QWidget *RegistrationListDelegate::createEditor(QWidget *parent, const QStyleOptionViewItem &,
                                                const QModelIndex &index) const
{
    auto *editor = new PlayersEditor(QString(), parent);
    Q_EMIT editorCreated(editor, index.data().toString());
    return editor;
}

void RegistrationListDelegate::updateEditorGeometry(QWidget *editor,
                                                    const QStyleOptionViewItem &option,
                                                    const QModelIndex &index) const
{
    int x = 2;
    int w = option.rect.width() - 2;
    if (index.data(Qt::DecorationRole).isValid()) {
        x += option.decorationSize.width();
        w -= option.decorationSize.width();
    }
    editor->setGeometry(x, option.rect.top() + 2,
                        w, option.rect.height() - 4);
}

void RegistrationListDelegate::setEditorData(QWidget *editor, const QModelIndex &index) const
{
    qobject_cast<PlayersEditor *>(editor)->setText(index.data().toString());
}

void RegistrationListDelegate::setModelData(QWidget *editor, QAbstractItemModel *,
                                            const QModelIndex &index) const
{
    const auto *playersEditor = qobject_cast<PlayersEditor *>(editor);

    if (! playersEditor->enterPressed()) {
        qCDebug(MuckturnierLog) << "Enter was not pressed, discarding edit";
        return;
    }

    const auto oldName = index.data().toString();
    const auto newName = playersEditor->text().simplified();
    if (newName == oldName) {
        qCDebug(MuckturnierLog) << "Name was not changed, discarding edit";
        return;
    }

    // Call RegistrationPage::renamePlayers() after everything is done here. Otherwise, we get
    // a segfault if renamePlayers() shows a message and a network change happens meanwhile
    QTimer::singleShot(0, [this, oldName, newName]
    {
        Q_EMIT requestRenamePlayers(oldName, newName);
    });
}

void RegistrationListDelegate::setHoveredRow(int row)
{
    m_hoveredRow = row;
}

void RegistrationListDelegate::setLastVisibleColumn(int column)
{
    m_lastVisibleColumn = column;
}

void RegistrationListDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option,
                                     const QModelIndex &index) const
{
    painter->save();

    const auto isNameColumn
        = index.column() == RegistrationColumns::columnIndex.value(RegistrationColumns::Name);
    auto opt = option;
    if (isNameColumn) {
        // Add two pixels to the width (which is the "selected" border with) so that we can move
        // the name two pixels to the right if the respective row is selected and we won't overlap
        // with the next column
        opt.rect.setWidth(option.rect.width() + 2);
    }

    const auto row = index.row();
    bool handledHere = false;

    if (row == m_hoveredRow) {
        handledHere = true;
        painter->fillRect(opt.rect, m_highlightBrush);
        if (m_hoveredRow > 0 && m_hoveredRow != selectedRow() + 1) {
            drawTopLine(painter, opt);
        }
    } else {
        drawBackground(painter, opt, index);
    }

    if (row == selectedRow()) {
        handledHere = true;

        painter->setPen(m_selectedPen);
        const auto column = index.column();

        if (column == 0) {
            // Left frame part
            painter->drawLine(QPoint(opt.rect.left() + 1, opt.rect.top() + 1),
                              QPoint(opt.rect.left() + 1, opt.rect.bottom()));
        }

        if (column == m_lastVisibleColumn) {
            // Right frame part
            painter->drawLine(QPoint(opt.rect.right(), opt.rect.top() + 1),
                              QPoint(opt.rect.right(), opt.rect.bottom()));
        }

        // Top frame part
        painter->drawLine(QPoint(opt.rect.left() + 1, opt.rect.top() + 1),
                          QPoint(opt.rect.right() - (column == m_lastVisibleColumn),
                                 opt.rect.top() + 1));

        // Bottom frame part
        painter->drawLine(QPoint(opt.rect.left() + 1, opt.rect.bottom()),
                          QPoint(opt.rect.right() - (column == m_lastVisibleColumn),
                                 opt.rect.bottom()));
    }

    painter->restore();

    if (handledHere) {
        if (index.row() == selectedRow() && isNameColumn) {
            // The row is selected and we draw the name column. So we move the text two pixels to
            // the right so that we have the same space between the frame and the text as we had
            // before between the widget border and the text. We compensate this by restoring the
            // original width (to which we added two pixels as a reserve above)
            opt.rect.setLeft(option.rect.left() + 2);
            opt.rect.setWidth(option.rect.width());
        }
        QStyledItemDelegate::paint(painter, opt, index);
    } else {
        TableDelegate::paint(painter, opt, index);
    }
}
