// SPDX-FileCopyrightText: 2010-2024 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "RegistrationPage.h"
#include "MarkersWidget.h"
#include "MarkersPopup.h"
#include "DrawPopup.h"
#include "DrawOverviewPage.h"
#include "DrawSettingsDialog.h"
#include "DisqualificationDialog.h"

#include "RegistrationListModel.h"
#include "RegistrationListView.h"
#include "RegistrationListDelegate.h"

#include "ServerPage/Server.h"
#include "ClientPage/Client.h"

#include "SharedObjects/SharedObjects.h"
#include "SharedObjects/StringTools.h"
#include "SharedObjects/SearchEngine.h"
#include "SharedObjects/Database.h"
#include "SharedObjects/Settings.h"
#include "SharedObjects/TournamentSettings.h"
#include "SharedObjects/DismissableMessage.h"

#include "shared/SearchWidget.h"
#include "shared/TitleMenu.h"
#include "shared/BackupEngine.h"
#include "shared/SharedStyles.h"
#include "shared/DefaultValues.h"
#include "shared/Tournament.h"
#include "shared/Draw.h"
#include "shared/PhoneticSearchButton.h"
#include "shared/ChecksumHelper.h"
#include "shared/Urls.h"
#include "shared/Players.h"
#include "shared/Logging.h"
#include "shared/IconButton.h"

// Qt includes
#include <QDebug>
#include <QVBoxLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QHeaderView>
#include <QScrollBar>
#include <QAction>
#include <QMessageBox>
#include <QApplication>
#include <QSplitter>
#include <QCheckBox>
#include <QDir>
#include <QTimer>
#include <QCursor>
#include <QPalette>

// C++ includes
#include <functional>
#include <utility>

static const QVector<QString> s_namePrefixes = {
    QStringLiteral("genannt"),
    QStringLiteral("von"),
    QStringLiteral("und"),
    QStringLiteral("zu"),
    QStringLiteral("van"),
    QStringLiteral("del"),
    QStringLiteral("de"),
    QStringLiteral("da"),
    QStringLiteral("d'"),
    QStringLiteral("d’")
};

static const QVector<QString> s_nameConnectors {
    QStringLiteral("-"),
    QStringLiteral("–"),
    QStringLiteral("'"),
    QStringLiteral("’")
};

static const QVector<QString> s_nameConnectorsReplacement {
    QStringLiteral("- "),
    QStringLiteral("– "),
    QStringLiteral("' "),
    QStringLiteral("’ ")
};

static const QLatin1Char s_space(' ');

RegistrationPage::RegistrationPage(QWidget *parent, SharedObjects *sharedObjects)
    : QWidget(parent),
      m_db(sharedObjects->database()),
      m_settings(sharedObjects->settings()),
      m_tournamentSettings(sharedObjects->tournamentSettings()),
      m_stringTools(sharedObjects->stringTools()),
      m_searchEngine(sharedObjects->searchEngine()),
      m_dismissableMessage(sharedObjects->dismissableMessage())
{
    // Markers
    m_markersWidget = new MarkersWidget(this, sharedObjects);
    connect(m_markersWidget, &MarkersWidget::markersChanged,
            this, &RegistrationPage::markersChanged);
    connect(m_markersWidget, &MarkersWidget::statusUpdate, this, &RegistrationPage::statusUpdate);
    connect(m_markersWidget, &MarkersWidget::visibilityChanged,
            this, std::bind(&RegistrationPage::updatePlayersList, this, QString(), -1));
    connect(m_markersWidget, &MarkersWidget::updatePlayersList,
            this, std::bind(&RegistrationPage::updatePlayersList, this, QString(), -1));

    QVBoxLayout *layout = new QVBoxLayout(this);

    // Register new players/pairs

    auto *registerWidget = new QWidget;
    auto *registerLayout = new QVBoxLayout(registerWidget);
    registerLayout->setContentsMargins(0, 0, 0, 0);

    m_newPlayersBox = new QGroupBox;
    m_newPlayersBox->setSizePolicy(QSizePolicy::Minimum, QSizePolicy::Maximum);
    auto *newPayersBoxWrapperLayout = new QVBoxLayout(m_newPlayersBox);
    auto *newPlayersBoxLayout = new QHBoxLayout;
    newPayersBoxWrapperLayout->addLayout(newPlayersBoxLayout);
    registerLayout->addWidget(m_newPlayersBox);

    m_nameLabel = new QLabel;
    newPlayersBoxLayout->addWidget(m_nameLabel);

    m_playersName = new QLineEdit;
    newPlayersBoxLayout->addWidget(m_playersName);
    connect(m_playersName, &QLineEdit::textChanged,
            [this](const QString &text)
            {
                m_registerButton->setEnabled(! text.simplified().isEmpty());
            });
    connect(m_playersName, &QLineEdit::returnPressed, this, &RegistrationPage::registerPlayers);
    connect(m_playersName, &QLineEdit::textChanged, this, &RegistrationPage::playersNameChanged);

    m_registerButton = new QPushButton(tr("Anmelden"));
    m_registerButton->setEnabled(false);
    newPlayersBoxLayout->addWidget(m_registerButton);
    connect(m_registerButton, &QPushButton::clicked, this, &RegistrationPage::registerPlayers);

    // Registration options

    auto *optionsLayout = new QHBoxLayout;
    optionsLayout->setContentsMargins(0, 0, 4, 0);
    newPayersBoxWrapperLayout->addLayout(optionsLayout);

    m_registerPhoneticSearch = new PhoneticSearchButton(Settings::registrationPhoneticId,
                                                        sharedObjects);
    connect(m_registerPhoneticSearch, &QAbstractButton::toggled,
            this, &RegistrationPage::registerPhoneticSearchToggled);
    optionsLayout->addWidget(m_registerPhoneticSearch);

    m_searchAsYouType = new IconButton(sharedObjects, QStringLiteral("search_as_you_type.png"),
                                       this);
    m_searchAsYouType->setToolTip(tr("<p>Beim Eingeben die Anmeldeliste durchsuchen</p>"));
    connect(m_searchAsYouType, &QAbstractButton::toggled,
            this, &RegistrationPage::searchAsYouTypeToggled);
    optionsLayout->addWidget(m_searchAsYouType);

    m_autoCapitalize = new IconButton(sharedObjects, QStringLiteral("capitalize.png"), this);
    m_autoCapitalize->setToolTip(tr("<p>Automatische Groß- und Kleinschreibung "
                                    "(de-)aktivierien</p>"));
    m_autoCapitalize->setChecked(m_settings->autoCapitalizeState());
    connect(m_autoCapitalize, &QAbstractButton::toggled,
            m_settings, &Settings::saveAutoCapitalizeState);
    optionsLayout->addWidget(m_autoCapitalize);

    m_searchAsYouTypeResult = new QLabel;
    m_searchAsYouTypeResult->setAlignment(Qt::AlignRight | Qt::AlignVCenter);
    optionsLayout->addWidget(m_searchAsYouTypeResult);

    // Registered players/pairs

    m_playersBox = new QGroupBox;
    QVBoxLayout *playersBoxLayout = new QVBoxLayout(m_playersBox);
    registerLayout->addWidget(m_playersBox);

    m_isSearchLabel = new QLabel;
    m_isSearchLabel->setAlignment(Qt::AlignCenter);
    m_isSearchLabel->setWordWrap(true);
    m_isSearchLabel->hide();
    playersBoxLayout->addWidget(m_isSearchLabel);

    m_visibilityDeviation = new QLabel;
    m_visibilityDeviation->setWordWrap(true);
    m_visibilityDeviation->setAlignment(Qt::AlignCenter);
    m_visibilityDeviation->setTextInteractionFlags(Qt::TextBrowserInteraction);
    connect(m_visibilityDeviation, &QLabel::linkActivated, this,
            &RegistrationPage::processStatusLinkClick);
    m_visibilityDeviation->hide();
    playersBoxLayout->addWidget(m_visibilityDeviation);

    m_listModel = new RegistrationListModel(this, sharedObjects, m_markersWidget->markersModel());

    m_listDelegate = new RegistrationListDelegate(this);

    m_listView = new RegistrationListView(sharedObjects);
    m_listView->setModel(m_listModel);
    m_listView->setItemDelegate(m_listDelegate);
    m_listView->setContextMenuPolicy(Qt::CustomContextMenu);

    connect(m_listView, &RegistrationListView::rowHovered,
            m_listDelegate, &RegistrationListDelegate::setHoveredRow);
    connect(m_listView, &RegistrationListView::rowSelected,
            m_listDelegate, &RegistrationListDelegate::setSelectedRow);
    connect(m_listDelegate, &RegistrationListDelegate::editorCreated,
            m_listView, &RegistrationListView::editorCreated);
    connect(m_listDelegate, &RegistrationListDelegate::requestRenamePlayers,
            this, &RegistrationPage::renamePlayers);
    connect(m_listView, &RegistrationListView::requestPopup,
            this, std::bind(&RegistrationPage::popupRequested, this,
                            std::placeholders::_1, QPoint(-1, -1)));
    connect(m_listModel, &QAbstractItemModel::dataChanged,
            m_listView, &QTableView::resizeColumnsToContents);
    connect(m_markersWidget, &MarkersWidget::visibilityChanged,
            this, &RegistrationPage::updateVisibility);
    connect(m_listView, &QTableView::customContextMenuRequested,
            this, &RegistrationPage::showContextMenu);

    m_listView->setColumnHidden(rc::columnIndex.value(rc::Column::Mark), true);
    m_listView->setColumnHidden(rc::columnIndex.value(rc::Column::Draw), true);
    m_listView->setColumnHidden(rc::columnIndex.value(rc::Column::Marker), true);

    auto *header = m_listView->horizontalHeader();
    header->setSectionsClickable(false);
    header->setContextMenuPolicy(Qt::CustomContextMenu);
    connect(header, &QHeaderView::customContextMenuRequested,
            this, &RegistrationPage::showHeaderContextMenu);

    playersBoxLayout->addWidget(m_listView);

    // Search widget

    auto *searchLayout = new QHBoxLayout;
    playersBoxLayout->addLayout(searchLayout);

    m_searchWidget = new SearchWidget(QStringLiteral("registration_page_search"), sharedObjects);
    connect(m_searchWidget, &SearchWidget::searchChanged, this, &RegistrationPage::updateSearch);
    searchLayout->addWidget(m_searchWidget);

    // Arrange the registering part and the markers part in a splitter

    QSplitter *splitter = new QSplitter(this);
    splitter->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Expanding);
    splitter->setHandleWidth(2);
    splitter->setStyleSheet(SharedStyles::visibleHorizontalSplitter);

    splitter->addWidget(registerWidget);
    splitter->setCollapsible(0, false);
    splitter->setStretchFactor(0, 100);

    splitter->addWidget(m_markersWidget);
    splitter->setCollapsible(1, false);
    splitter->setStretchFactor(1, 1);

    layout->addWidget(splitter);

    // Wrapper widget for the status label and the options settings menu button
    auto *statusWidget = new QWidget;
    statusWidget->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    auto *statusLayout = new QHBoxLayout(statusWidget);
    statusLayout->setContentsMargins(0, 0, 0, 0);
    layout->addWidget(statusWidget);

    // Status label
    m_statusLabel = new QLabel;
    m_statusLabel->setWordWrap(true);
    m_statusLabel->setTextInteractionFlags(Qt::TextBrowserInteraction);
    m_statusLabel->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    connect(m_statusLabel, &QLabel::linkActivated, this, &RegistrationPage::processStatusLinkClick);
    statusLayout->addWidget(m_statusLabel);

    // Settings menu button

    m_menuButton = new QPushButton(tr("Einstellungen"));
    m_menuButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);
    auto *settingsMenu = new QMenu(this);
    m_menuButton->setMenu(settingsMenu);

    m_columnsMenu = new TitleMenu(tr("Angezeigte Spalten"), this);
    m_columnsMenu->setTitleHidden(true);

    for (const rc::Column column : rc::allColumns) {
        m_showColumn.insert(column, m_columnsMenu->addAction(tr(rc::checkBoxText.value(column))));
        m_showColumn[column]->setCheckable(true);
        connect(m_showColumn.value(column), &QAction::toggled,
                this, std::bind(&RegistrationPage::setColumnVisible, this, column));
    }

    m_columnsMenu->addSeparator();
    m_hideHeaderAction = m_columnsMenu->addAction(tr("Kopfzeile ausblenden"));
    m_hideHeaderAction->setCheckable(true);
    connect(m_hideHeaderAction, &QAction::toggled, this, &RegistrationPage::setHideHeader);

    settingsMenu->addMenu(m_columnsMenu);

    settingsMenu->addSeparator();
    auto *showDrawSettingsAction = settingsMenu->addAction(tr("Einstellungen für die Auslosung"));
    connect(showDrawSettingsAction, &QAction::triggered, this, &RegistrationPage::showDrawSettings);

    auto *menuButtonLayout = new QVBoxLayout;
    statusLayout->addLayout(menuButtonLayout);
    menuButtonLayout->addStretch();
    menuButtonLayout->addWidget(m_menuButton);

    // Markers popup
    m_markersPopup = new MarkersPopup(this, m_markersWidget->markers());
    connect(m_markersPopup, &QDialog::accepted,
            [this]
            {
                updateMarker(m_markersPopup->markerName(), m_markersPopup->markerId());
            });

    // Draw popup
    m_drawPopup = new DrawPopup(this, &m_tournamentSettings->drawSettings(),
                                m_listModel->assignedTables(1), m_listModel->assignedTables(2));
    connect(m_listModel, &QAbstractItemModel::dataChanged,
            m_drawPopup, &DrawPopup::updateFirstFreeTable);
    connect(m_drawPopup, &QDialog::accepted,
            [this]
            {
                updateDraw(m_drawPopup->seat());
            });
    // This is needed if another network participant sets a draw and the popup is visible
    connect(this, &RegistrationPage::drawChanged, m_drawPopup, &DrawPopup::checkSelection);

    // Context menu

    m_contextMenu = new TitleMenu(this);
    connect(m_contextMenu, &QMenu::aboutToHide,
            [this]
            {
                m_lastContextMenuPos = m_contextMenu->geometry().topLeft();
            });

    m_assignMenu = m_contextMenu->addMenu(tr("Einzelspieler zuordnen"));
    m_assignMenu->menuAction()->setVisible(false);

    m_editAction = m_contextMenu->addAction(QString());
    connect(m_editAction, &QAction::triggered,
            m_listView, &RegistrationListView::editCurrentName);

    m_deleteAction = m_contextMenu->addAction(QString());
    connect(m_deleteAction, &QAction::triggered, this, &RegistrationPage::deletePlayers);

    m_contextMenu->addSeparator();

    m_setOnlyMarkerAction = m_contextMenu->addAction(QString());
    m_setOnlyMarkerAction->setVisible(false);

    m_setMarkerMenu = m_contextMenu->addMenu(tr("Markieren als"));

    m_removeMarkerAction = m_contextMenu->addAction(tr("Markierung entfernen"));
    connect(m_removeMarkerAction, &QAction::triggered,
            this, std::bind(&RegistrationPage::updateMarker, this, QString(), 0));

    m_contextMenu->addSeparator();

    m_setDrawAction = m_contextMenu->addAction(tr("Auslosung eingeben"));
    connect(m_setDrawAction, &QAction::triggered,
            [this]
            {
                popupRequested(rc::Draw, m_lastContextMenuPos);
            });

    m_deleteDrawAction = m_contextMenu->addAction(tr("Auslosung löschen"));
    connect(m_deleteDrawAction, &QAction::triggered,
            this, std::bind(&RegistrationPage::updateDraw, this, Draw::Seat { 0, 0 }));

    // Disqualify menu

    m_contextMenu->addSeparator();
    m_disqualificationMenu = m_contextMenu->addMenu(tr("Disqualifikation"));

    m_disqualifyAction = m_disqualificationMenu->addAction(tr("Disqualifizieren"));
    connect(m_disqualifyAction, &QAction::triggered, this, &RegistrationPage::setDisqualified);

    m_unDisqualifyAction = m_disqualificationMenu->addAction(tr("Disqualifikation zurücknehmen"));
    connect(m_unDisqualifyAction, &QAction::triggered, this, &RegistrationPage::setDisqualified);

    // Whole list menu

    m_listMenu = new TitleMenu(tr("Ganze Liste"), this);

    m_contextMenu->addSeparator();
    m_contextMenu->addMenu(m_listMenu);

    m_listSetOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listSetMarkerMenu = m_listMenu->addMenu(tr("Markieren als"));

    m_listRemoveOnlyMarkerAction = m_listMenu->addAction(QString());

    m_listRemoveMarkerMenu = m_listMenu->addMenu(tr("Markierungen entfernen"));

    m_listRemoveAllMarkersAction = m_listMenu->addAction(tr("Alle Markierungen entfernen"));
    connect(m_listRemoveAllMarkersAction, &QAction::triggered,
            this, std::bind(&RegistrationPage::listUpdateMarker, this, QString(), 0));

    m_listMenu->addSeparator();

    m_deleteOnlyMarkedAction = m_listMenu->addAction(QString());

    m_deleteMarkedMenu = m_listMenu->addMenu(QString());

    m_deleteUnmarkedAction = m_listMenu->addAction(QString());
    connect(m_deleteUnmarkedAction, &QAction::triggered, this, &RegistrationPage::deleteUnmarked);

    m_deleteAllAction = m_listMenu->addAction(QString());
    connect(m_deleteAllAction, &QAction::triggered, this, &RegistrationPage::deleteAll);

    m_listMenu->addSeparator();

    auto *drawAllSeatsAction = m_listMenu->addAction(tr("Alle Plätze auslosen"));
    connect(drawAllSeatsAction, &QAction::triggered, this, &RegistrationPage::drawAllSeats);

    m_deleteAllDrawsAction = m_listMenu->addAction(tr("Alle Auslosungen löschen"));
    connect(m_deleteAllDrawsAction, &QAction::triggered, this, &RegistrationPage::deleteAllDraws);

    // Set the m_registerPhoneticSearch button's tool tip and style sheet
    registerPhoneticSearchToggled();

    // Restore saved settings
    m_searchAsYouType->setChecked(m_settings->searchAsYouTypeState());
}

void RegistrationPage::reload(bool updateMarkersWidget)
{
    m_newPlayersBox->show();

    switch (m_tournamentSettings->tournamentMode()) {
    case Tournament::FixedPairs:
        m_newPlayersBox->setTitle(tr("Neue Paare anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Paare"));
        m_nameLabel->setText(tr("Name(n):"));
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAction->setText(tr("Paar löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Paare mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Paare löschen"));
        m_deleteAllAction->setText(tr("Alle Paare löschen"));
        break;
    case Tournament::SinglePlayers:
        m_newPlayersBox->setTitle(tr("Neue Spieler anmelden"));
        m_playersBox->setTitle(tr("Angemeldete Spieler"));
        m_nameLabel->setText(tr("Name:"));
        m_editAction->setText(tr("Name ändern"));
        m_deleteAction->setText(tr("Spieler löschen"));
        m_deleteMarkedMenu->menuAction()->setText(tr("Alle Spieler mit Markierung löschen"));
        m_deleteUnmarkedAction->setText(tr("Alle unmarkierten Spieler löschen"));
        m_deleteAllAction->setText(tr("Alle Spieler löschen"));
        break;
    }

    m_newPlayersBox->setEnabled(m_db->isChangeable());
    if (! m_db->isChangeable()) {
        m_listView->deselectRow();
    }
    m_listView->setEditTriggers(m_db->isChangeable() ? QAbstractItemView::DoubleClicked
                                                     : QAbstractItemView::NoEditTriggers);
    m_drawPopup->setPairMode(m_tournamentSettings->isPairMode());
    m_markersWidget->updateTournamentMode();

    // We have to call this via a QTimer call if the database is opened via command line argument
    // to be sure that everything is constructed and set up before we do this
    QTimer::singleShot(0, [this]
    {
        m_hideHeaderAction->setChecked(m_tournamentSettings->registrationListHeaderHidden());
    });

    // This will also trigger updatePlayersList() each
    if (updateMarkersWidget) {
        m_markersWidget->reload();
    } else {
        markersChanged();
    }

    checkUpdateDrawOverviewPage(true);
}

void RegistrationPage::markersChanged()
{
    const auto &markers = m_markersWidget->markers();
    const int count = markers.count();

    // No need to process count < 2, because the actions will be hidden in this case

    if (count == 2) {
        // We have exactly one marker

        m_setOnlyMarkerAction->disconnect();
        m_listSetOnlyMarkerAction->disconnect();
        m_listRemoveOnlyMarkerAction->disconnect();
        m_deleteOnlyMarkedAction->disconnect();

        for (const Markers::Marker &marker : markers) {
            if (marker.id != 0) {
                m_setOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_setOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::updateMarker, this,
                                        marker.name, marker.id));

                m_listSetOnlyMarkerAction->setText(tr("Markieren als „%1“").arg(marker.name));
                connect(m_listSetOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::listUpdateMarker, this,
                                        marker.name, marker.id));

                m_listRemoveOnlyMarkerAction->setText(tr("Markierung „%1“ entfernen").arg(
                                                          marker.name));
                connect(m_listRemoveOnlyMarkerAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::listRemoveMarker,
                                        this, marker.name, marker.id));

                m_deleteOnlyMarkedAction->setText((m_tournamentSettings->isPairMode()
                    ? tr("Alle als „%1“ markierten Paare löschen")
                    : tr("Alle als „%1“ markierten Spieler löschen")).arg(marker.name));
                m_deleteOnlyMarkedAction->setData(marker.id);
                connect(m_deleteOnlyMarkedAction, &QAction::triggered,
                        this, std::bind(&RegistrationPage::deleteMarked, this,
                                        marker.name, marker.id));

                break;
            }
        }
    }

    if (count > 2) {
        // We have multiple markers
        m_setMarkerMenu->clear();
        m_listSetMarkerMenu->clear();
        m_listRemoveMarkerMenu->clear();
        m_deleteMarkedMenu->clear();

        for (const Markers::Marker &marker : markers) {
            if (marker.id == 0) {
                continue;
            }

            QAction *setMarkerAction = m_setMarkerMenu->addAction(marker.name);
            setMarkerAction->setData(marker.id);
            connect(setMarkerAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::updateMarker, this,
                                    marker.name, marker.id));

            QAction *listSetMarkerAction = m_listSetMarkerMenu->addAction(marker.name);
            listSetMarkerAction->setData(marker.id);
            connect(listSetMarkerAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::listUpdateMarker, this,
                                    marker.name, marker.id));

            QAction *listRemoveMarkerAction = m_listRemoveMarkerMenu->addAction(marker.name);
            listRemoveMarkerAction->setData(marker.id);
            connect(listRemoveMarkerAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::listRemoveMarker, this,
                                    marker.name, marker.id));

            QAction *deleteMarkedAction = m_deleteMarkedMenu->addAction(marker.name);
            deleteMarkedAction->setData(marker.id);
            connect(deleteMarkedAction, &QAction::triggered,
                    this, std::bind(&RegistrationPage::deleteMarked, this,
                                    marker.name, marker.id));
        }
    }

    m_setOnlyMarkerAction->setVisible(count == 2);
    m_setMarkerMenu->menuAction()->setVisible(count > 2);
    m_removeMarkerAction->setVisible(count > 1);

    m_listSetOnlyMarkerAction->setVisible(count == 2);
    m_listSetMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveOnlyMarkerAction->setVisible(count == 2);
    m_listRemoveMarkerMenu->menuAction()->setVisible(count > 2);
    m_listRemoveAllMarkersAction->setVisible(count > 2);

    m_deleteOnlyMarkedAction->setVisible(count == 2);
    m_deleteMarkedMenu->menuAction()->setVisible(count > 2);
    m_deleteUnmarkedAction->setVisible(count > 2);

    const auto selectedName = m_listView->selectedName();
    updatePlayersList();
    if (! selectedName.isEmpty()) {
        m_listView->selectName(selectedName);
    }

    // Update the markers popup
    m_markersPopup->updateMarkers();
}

void RegistrationPage::updatePlayersList(const QString &changedSelection, int originalVisibleRow)
{
    // This is only for a possibly existing draw adjustment
    // dialog if we just processed a network change.
    Q_EMIT listWasModified();

    const auto oldRegistrationsCount = m_listModel->rowCount();
    m_listModel->refresh();
    if (! changedSelection.isEmpty()) {
        m_listView->selectName(changedSelection, originalVisibleRow);
    }
    const auto registrationsCount = m_listModel->rowCount();

    const bool drawsSet = m_listModel->assignmentPresent();
    m_deleteAllDrawsAction->setEnabled(drawsSet);
    updateVisibility();

    // Update the context menus

    if (m_listSetMarkerMenu->menuAction()->isVisible()) {
        const auto listSetMarkerMenuActions = m_listSetMarkerMenu->actions();
        for (QAction *action : listSetMarkerMenuActions) {
            action->setEnabled(m_listModel->markerCount(action->data().toInt())
                                   < registrationsCount);
        }
        const auto listRemoveMarkerMenuActions = m_listRemoveMarkerMenu->actions();
        for (QAction *action : listRemoveMarkerMenuActions) {
            action->setEnabled(m_listModel->markerCount(action->data().toInt()) > 0);
        }
        const auto deleteMarkedMenuActions = m_deleteMarkedMenu->actions();
        for (QAction *action : deleteMarkedMenuActions) {
            action->setEnabled(m_listModel->markerCount(action->data().toInt()) > 0);
        }
    }

    if (m_listModel->markerCount(0) == registrationsCount) {
        m_listRemoveOnlyMarkerAction->setEnabled(false);
        m_listRemoveMarkerMenu->setEnabled(false);
        m_listRemoveAllMarkersAction->setEnabled(false);
    } else {
        m_listRemoveOnlyMarkerAction->setEnabled(true);
        m_listRemoveMarkerMenu->setEnabled(true);
        m_listRemoveAllMarkersAction->setEnabled(true);
    }

    if (m_db->tournamentStarted() || m_listModel->markerCount(0) == registrationsCount) {
        m_deleteOnlyMarkedAction->setEnabled(false);
        m_deleteMarkedMenu->setEnabled(false);
        m_deleteUnmarkedAction->setEnabled(false);
    } else {
        m_deleteOnlyMarkedAction->setEnabled(
            m_listModel->markerCount(m_deleteOnlyMarkedAction->data().toInt())
                < registrationsCount);
        m_deleteMarkedMenu->setEnabled(true);
        m_deleteUnmarkedAction->setEnabled(m_listModel->markerCount(0) > 0);
    }

    // Update the GUI

    m_markersWidget->updateCounters(m_listModel->markerCounts());

    m_searchWidget->setVisible(registrationsCount > 0
                               && (! m_searchAsYouType->isChecked() || m_db->tournamentStarted()));
    Q_EMIT registrationsAvailable(registrationsCount > 0);

    const int singleMarker = m_db->tournamentStarted() ? -1 : m_markersWidget->singleMarker();
    const bool oldSinglesPresent = m_singlesPresent;
    m_singlesPresent = m_listModel->markerCount(singleMarker) > 0;
    const bool singlesPresentChanged = oldSinglesPresent != m_singlesPresent;

    const int absentMarker = m_db->tournamentStarted()
        ? -1 : m_tournamentSettings->marker(Markers::Absent);
    const bool oldAbsentsPresent = m_absentsPresent;
    m_absentsPresent = m_listModel->markerCount(absentMarker) > 0;
    const bool absentsPresentChanged = oldAbsentsPresent != m_absentsPresent;

    int presentPairsWithoutSingles = registrationsCount;
    if (singleMarker != -1) {
        presentPairsWithoutSingles -= m_listModel->markerCount(singleMarker);
    }
    if (absentMarker != -1) {
        presentPairsWithoutSingles -= m_listModel->markerCount(absentMarker);
    }

    // Update the registration status

    m_drawAdjustmentNeeded = false;
    QString registrationStatus;

    if (m_absentsPresent) {
        registrationStatus.append(tr("<span style=\"%1\">%n Anmeldung(en) ignoriert</span> – ", "",
                                     m_listModel->markerCount(absentMarker)).arg(
                                         SharedStyles::redText));
    }

    registrationStatus.append(m_tournamentSettings->isPairMode()
                                  ? tr("%n Paar(e)", "", presentPairsWithoutSingles)
                                  : tr("%n Spieler", "", presentPairsWithoutSingles));

    if (! m_singlesPresent) {
        // All registrations are valid and can be counted as-is
        registrationStatus.append(tr(" angemeldet"));

        if (presentPairsWithoutSingles > 0) {
            if (m_db->playersDividable(presentPairsWithoutSingles)) {
                registrationStatus.append(takenTablesString(
                    presentPairsWithoutSingles / m_db->necessaryDivisor(), true));
            } else {
                registrationStatus.append(m_tournamentSettings->isPairMode()
                    ? tr(" – <span style=\"%1\">ein Paar fehlt noch</span>").arg(
                         SharedStyles::redText)
                    : tr(" – <span style=\"%1\">%n Spieler fehlt/fehlen noch</span>", "",
                         m_db->necessaryDivisor()
                         - presentPairsWithoutSingles % m_db->necessaryDivisor()).arg(
                         SharedStyles::redText));
                registrationStatus.append(takenTablesString(
                    presentPairsWithoutSingles / m_db->necessaryDivisor() + 1, false));
            }
        }

    } else {
        // We're in pair mode and have to consider single players
        const int remainder = (presentPairsWithoutSingles * 2
                               + m_listModel->markerCount(singleMarker)) % 4;
        registrationStatus.append(tr(" und <span style=\"%1\"> %n allein gekommene(r) Spieler"
                                     "</span> angemeldet<br/>", "",
                                     m_listModel->markerCount(singleMarker)).arg(
                                         SharedStyles::redText));
        if (remainder == 0) {
            registrationStatus.append(tr("<span style=\"%1\">Wenn alle zugeordnet sind, kann es "
                                         "losgehen</span>").arg(SharedStyles::greenText));
        } else if (remainder == 1) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein Paar und ein "
                                         "einzelner Spieler</span>").arg(
                                         SharedStyles::redText));
        } else if (remainder == 2) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein Paar</span>").arg(
                                         SharedStyles::redText));
        } else if (remainder == 3) {
            registrationStatus.append(tr("<span style=\"%1\">Es fehlt noch ein einzelner Spieler"
                                         "</span>").arg(SharedStyles::redText));
        }
        registrationStatus.append(takenTablesString(
            (presentPairsWithoutSingles * 2 + m_listModel->markerCount(singleMarker)) / 4
            + (remainder != 0), false));
    }

    // Add the "no draw yet" line if we have a draw

    if (m_showColumn.value(rc::Column::Draw)->isChecked() && ! drawsSet) {
        registrationStatus.append(tr("<br/>Es wurde noch keine Auslosung eingegeben"));

    } else if (drawsSet) {
        const auto incompleteTables = m_listModel->incompleteTables().count() > 0;
        const auto tableGaps = m_listModel->tableGaps().count() > 0;

        if (m_drawOverviewPage == nullptr) {
            registrationStatus.append(tr("<br/>Auslosung (<a href=\"#drawOverview\">Übersicht"
                                         "</a>): "));
        } else {
            registrationStatus.append(tr("<br/>Auslosung: "));
        }

        const bool unfinished = m_listModel->drawsCount() != registrationsCount;

        // Append the draw status to the status line

        registrationStatus.append(unfinished
            ? tr("<a href=\"#unfinished\">Unvollständig</a>; ")
            : tr("Vollständig; "));

        if (incompleteTables) {
            registrationStatus.append(tr("<a href=\"#incompleteTables\">unvollständige Tische"
                                         "</a>"));
        }

        if (tableGaps) {
            registrationStatus.append(incompleteTables
                ? tr("; <a href=\"#tableGaps\">nicht fortlaufend</a>")
                : tr("<a href=\"#tableGaps\">nicht fortlaufend</a>"));
        }

        if (! incompleteTables && ! tableGaps) {
            registrationStatus.append(tr("alle Tische sind fortlaufend und vollständig vergeben"));
        }

        m_drawAdjustmentNeeded = ! unfinished && ! m_singlesPresent && ! m_absentsPresent
                                 && m_db->playersDividable(presentPairsWithoutSingles)
                                 && (incompleteTables || tableGaps);
        if (m_drawAdjustmentNeeded) {
            registrationStatus.append(tr(" – <a href=\"#adjustDraw\">Auslosung bereinigen</a>"));
        }
    }

    Q_EMIT drawAdjustmentNeeded(m_drawAdjustmentNeeded);

    // Finish the update

    m_statusLabel->setText(registrationStatus);

    if (registrationsCount != oldRegistrationsCount
        || singlesPresentChanged || absentsPresentChanged) {

        m_db->setPairsOrPlayersCount(registrationsCount);
        // Emitting the registrationsCountChanged signal will make the MainWindow check for single
        // players being present and hide/show the score dock accordingly, also if the number of
        // registered players wasn't changed by the previous action (e. g. marking an entry as
        // single)
        Q_EMIT playersCountChanged(registrationsCount);
    }

    if (m_db->tournamentStarted()) {
        m_newPlayersBox->hide();
    }

    scoreChanged();

    if (m_isServer || m_isClient) {
        updateChecksum();
    }
}

QString RegistrationPage::takenTablesString(int tables, bool definitive) const
{
    const int tablesLimit = m_tournamentSettings->drawSettings().tablesLimit;

    if (tablesLimit == -1 || tables <= tablesLimit) {
        if (definitive) {
            return tr(" (%n Tisch(e) besetzt)", "", tables);
        } else {
            return tr(" (dann wäre(n) %n Tisch(e) besetzt)", "", tables);
        }
    } else {
        if (definitive) {
            return tr(" (%n Tisch(e) besetzt – <span style=\"%1\">Limit laut "
                      "Auslosungseinstellungen: %2</span>)", "", tables).arg(
                      SharedStyles::redText, QString::number(tablesLimit));
        } else {
            return tr(" (dann wäre(n) %n Tisch(e) besetzt – <span style=\"%1\">Limit laut "
                      "Auslosungseinstellungen: %2</span>)", "", tables).arg(
                      SharedStyles::redText, QString::number(tablesLimit));
        }
    }
}

void RegistrationPage::scoreChanged()
{
    m_deleteAction->setEnabled(! m_db->tournamentStarted());
    m_deleteAllAction->setEnabled(! m_db->tournamentStarted());
    m_markersWidget->updateTournamentState();
}

void RegistrationPage::reset()
{
    m_listView->deselectRow();
    m_listModel->clear();
    m_newPlayersBox->show();
}

bool RegistrationPage::checkForMarkAsAboard()
{
    // Check if an absent marker is defined
    const auto originalAbsentId = m_tournamentSettings->marker(Markers::Absent);
    if (originalAbsentId == -1) {
        return false;
    }

    // Check if a possible current search only shows one match
    int onlyVisibleIndex = -1;
    for (int i = 0; i < m_listModel->rowCount(); i++) {
        if (m_actuallyHiddenRow == i) {
            // This row would normally be hidden by the search and is only
            // visible because it's currently selected. Thus, we skip it.
            continue;
        }
        if (! m_listView->isRowHidden(i)) {
            if (onlyVisibleIndex != -1) {
                return false;
            }
            onlyVisibleIndex = i;
        }
    }
    if (onlyVisibleIndex == -1) {
        return false;
    }

    // Check if that one match is currently marked as absent
    const auto playersData = m_listModel->data(onlyVisibleIndex);
    if (playersData.marker != originalAbsentId) {
        return false;
    }

    const auto originalAbsentMarkerData = m_markersWidget->markerData(originalAbsentId);
    const auto defaultMarkerData = m_markersWidget->markerData(
        m_tournamentSettings->marker(Markers::Default));

    if (QMessageBox::question(this,
            tr("Voranmeldung als „gekommen“ markieren"),
            tr("<p>Die Eingabe „%1“ passt exakt zu einer momentan ignorierten Anmeldung "
               "(höchstwahrscheinlich eine Voranmeldung), nämlich</p>"
               "<p>„%2“</p>"
               "<p>Soll %3 als „%4“ markiert werden?</p>").arg(
               m_playersName->text().simplified(),
               playersData.name,
               m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
               defaultMarkerData.name),
               QMessageBox::Yes | QMessageBox::No)
        != QMessageBox::Yes) {

        return false;
    }

    if (m_isClient || m_isServer) {
        // We have to check for possible colliding network changes

        // Check for a name change or deletion
        if (! m_listModel->hasExactMatch(playersData.name)) {
            QTimer::singleShot(0, [this, playersData]
            {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("<p>%1 „%2“ wurde zwischenzeitlich von einem anderen Netzwerkteilnehmer "
                       "entweder umbenannt oder gelöscht. Die Änderung der Markierung ist nicht "
                       "mehr möglich.</p>"
                       "<p>Bitte die Änderung erneut anfragen!</p>").arg(
                       m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                       playersData.name));
            });
            return true;
        }

        // Check for a change of the "default marker" or "absent marker" settings.
        // The "default marker" can't be changed via network, so we don't have to check it.

        if (originalAbsentId != m_tournamentSettings->marker(Markers::Absent)
            || (originalAbsentId != -1
                && originalAbsentMarkerData
                    != m_markersWidget->markerData(
                           m_tournamentSettings->marker(Markers::Absent)))) {

            QTimer::singleShot(0, [this]
            {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("<p>Die Einstellung für die Markierung neuer bzw. abwesender Anmeldungen "
                       "wurde(n) zwischenzeitlich von einem anderen Netzwerkteilnehmer geändert. "
                       "Die Änderung der Markierung ist nicht mehr möglich.</p>"
                       "<p>Bitte die Änderung erneut anfragen!</p>"));
            });

            return true;
        }
    }

    m_playersName->clear();
    m_listView->selectName(playersData.name);
    updateMarker(defaultMarkerData.name, defaultMarkerData.id);

    if (m_isServer) {
        // Be sure to scroll to the changed entry. In client mode, the change is
        // processed by processNetworkChange() anyway, and the scrolling is ensured.
        QTimer::singleShot(0,
            [this, playersData, onlyVisibleIndex]
            {
                m_listView->selectName(playersData.name, onlyVisibleIndex);
            });
    }

    return true;
}

void RegistrationPage::registerPlayers()
{
    if (! m_registerButton->isEnabled()) {
        return;
    }

    // If we have an absent marker defined, check if a possible search currently only
    // shows one match, and if that one is marked as absent, ask to change the marker
    if (checkForMarkAsAboard()) {
        return;
    }

    QString name = m_playersName->text().simplified();

    // If requested, auto-capitalize the name(s)

    if (m_autoCapitalize->isChecked()) {
        for (int i = 0; i < s_nameConnectors.count(); i++) {
            name = name.replace(s_nameConnectors.at(i), s_nameConnectorsReplacement.at(i));
        }

        const auto &namesSeparator = m_settings->namesSeparator();
        const auto trimmedNamesSeparator = namesSeparator.trimmed();
        const auto namesSeparatorWithSpaces = QStringLiteral(" %1 ").arg(trimmedNamesSeparator);
        name = name.replace(trimmedNamesSeparator, namesSeparatorWithSpaces);

        const auto nameParts = name.split(s_space, Qt::SkipEmptyParts);

        QStringList processedNameParts;

        for (const auto &part : nameParts) {
            const auto lowerPart = m_stringTools->toLower(part);
            bool isNamePrefix = false;
            for (const auto &prefix : s_namePrefixes) {
                if (lowerPart == prefix) {
                    isNamePrefix = true;
                    break;
                }
            }

            if (isNamePrefix) {
                processedNameParts.append(part);
            } else {
                QString upperCasePart = part;
                upperCasePart.replace(0, 1, m_stringTools->toUpper(QString(upperCasePart[0])));
                processedNameParts.append(upperCasePart);
            }
        }

        name = processedNameParts.join(s_space);
        for (int i = 0; i < s_nameConnectors.count(); i++) {
            name = name.replace(s_nameConnectorsReplacement.at(i), s_nameConnectors.at(i));
        }

        name = name.replace(namesSeparatorWithSpaces, namesSeparator);
    }

    // Check for an exact match. In this case, we can't register the pair/player
    if (m_listModel->hasCaseInsensitiveMatch(name)) {
        QMessageBox::warning(this, m_newPlayersBox->title(),
            tr("<p><b>%1 „%2“ kann nicht angemeldet werden</b></p>"
               "<p>Der Name ist bereits vergeben.</p>").arg(
               m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"), name));
        m_playersName->selectAll();
        m_playersName->setFocus();
        return;
    }

    // Check for a fuzzy match. If one is found, display a warning about a possible duplicate
    const auto fuzzyMatch = m_listModel->hasFuzzyMatch(name, m_registerPhoneticSearch->isChecked());
    if (! fuzzyMatch.isEmpty()
        && QMessageBox::warning(this,
            m_newPlayersBox->title(),
            tr("<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
               "<p><table>"
               "<tr><td>Eingegeben:&nbsp;</td><td><i>%2</i></td></tr>"
               "<tr><td>Vorhanden:</td><td><i>%3</i></td></tr>"
               "</table></p>"
               "<p>Soll %4 trotzdem angemeldet werden?</p>").arg(
               m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
               name,
               fuzzyMatch,
               m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("der Spieler")),
               QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
        == QMessageBox::No) {

           m_playersName->selectAll();
           m_playersName->setFocus();
           return;
    }

    if (m_isClient || m_isServer) {
        // Someone else could have registered the exact name we want to register meanwhile
        if (m_listModel->hasCaseInsensitiveMatch(name)) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer ein %1 mit dem "
                   "Namen „%2“ angemeldet. Die Anmeldung ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"), name));
            m_playersName->clear();
            return;
        }
    }

    if (m_isClient) {
        m_requestingRegistration = true;
        m_client->requestRegisterPlayers(name, m_markersWidget->newRegistrationMarker(name));
        m_focusPlayersName = true;
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(m_tournamentSettings->isPairMode() ? tr("Melde Paar an …")
                                                           : tr("Melde Spieler an …"));

    if (! m_db->registerPlayers(name, m_markersWidget->newRegistrationMarker(name))) {
        m_playersName->clear();
        return;
    }

    m_playersName->clear();
    m_playersName->setFocus();
    updatePlayersList();
    if (! m_searchAsYouType->isChecked()) {
        m_searchWidget->clear();
    }

    QTimer::singleShot(0, [this, name]
    {
        m_listView->selectName(name);
    });

    Q_EMIT statusUpdate((m_tournamentSettings->isPairMode() ? tr("Paar „%1“ angemeldet")
                                                            : tr("Spieler „%1“ angemeldet")).arg(
                                                                 name));

    if (m_isServer) {
        m_server->broadcastRegisterPlayers(name, m_markersWidget->newRegistrationMarker(name));
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::renamePlayers(const QString &originalName, const QString &name)
{
    // Check for an existing registration with the name we want to rename to.
    // In this case, we can't rename the pair/player.

    const auto pureCaseChange = m_stringTools->toLower(originalName)
                                    == m_stringTools->toLower(name);

    if (! pureCaseChange && m_listModel->hasCaseInsensitiveMatch(name)) {
        QMessageBox::warning(this,
            m_tournamentSettings->isPairMode() ? tr("Paar umbenennen") : tr("Spieler umbenennen"),
            tr("<p><b>%1 „%2“ kann nicht umbenannt werden:</b></p>"
               "<p>Der Name „%3“ ist bereits vergeben.</p>").arg(
               m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
               originalName, name));
        return;
    }

    // Seach for a fuzzy match

    const auto fuzzyMatch = m_listModel->hasFuzzyMatch(name,
                                                       m_registerPhoneticSearch->isChecked(),
                                                       originalName);

    if (! fuzzyMatch.isEmpty()) {
        m_fuzzyMatchWarning = new QMessageBox(QMessageBox::Warning,
            m_tournamentSettings->isPairMode() ? tr("Paar umbenennen") : tr("Spieler umbenennen"),
            tr("<p>Umbenennen von %1 „%2“</p>"
               "<p>Es ist bereits ein %1 mit einem ähnlichen Namen angemeldet:</p>"
               "<p><table>"
               "<tr><td>Eingegeben:&nbsp;</td><td><i>%3</i></td></tr>"
               "<tr><td>Vorhanden:</td><td><i>%4</i></td></tr>"
               "</table></p>"
               "<p>Das könnte zu Verwechslungen führen. Soll %5 trotzdem umbenannt "
               "werden?</p>").arg(
               m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
               originalName,
               name,
               fuzzyMatch,
               m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("Der Spieler")),
            QMessageBox::Yes | QMessageBox::No,
            this);
        m_fuzzyMatchWarning->setDefaultButton(QMessageBox::No);

        bool abort = ! m_fuzzyMatchWarning->exec()
                     || m_fuzzyMatchWarning->clickedButton()
                        != m_fuzzyMatchWarning->button(QMessageBox::Yes);
        m_fuzzyMatchWarning->deleteLater();

        if (abort) {
            return;
        }
    }

    // Check for colliding network changes

    if (m_isClient || m_isServer) {
        // Someone could have deleted or renamed or edited entry
        if (m_listModel->rowForName(originalName) == -1) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "umbenannt oder gelöscht. Das Umbenennen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                   originalName));
            return;
        }

        // Someone could have registered a new pair or player with the exact name we want to rename
        // our edited pair or player to or could have renamed another one to that exact name
        if (! pureCaseChange && m_listModel->hasCaseInsensitiveMatch(name)) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("Es wurde mittlerweile von einem anderen Netzwerkteilnehmer %1 mit dem "
                   "Namen „%2“ angemeldet oder %1 entsprechend umbenannt. Das Umbenennen ist "
                   "nicht mehr möglich und wird abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("ein Paar") : tr("ein Spieler"), name));
            return;
        }
    }

    const auto [ markerAfterEditId, markerAfterEditName ]
        = m_markersWidget->markerAfterEdit(originalName, name);

    if (m_isClient) {
        m_client->requestRenamePlayers(originalName, name, markerAfterEditId, markerAfterEditName);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    QString changeMarkerStartMessage;
    QString changeMarkerFinishedMessage;
    if (markerAfterEditId != -1) {
        if (markerAfterEditId == 0) {
            changeMarkerStartMessage = tr("und entferne die Markierung ");
            changeMarkerFinishedMessage = tr("und die Markierung entfernt");
        } else {
            changeMarkerStartMessage = tr(" und setze die Markierung auf „%1“ ").arg(
                                          markerAfterEditName);
            changeMarkerFinishedMessage = tr(" und die Markierung auf „%1“ gesetzt").arg(
                                             markerAfterEditName);
        }
    }

    Q_EMIT statusUpdate((m_tournamentSettings->isPairMode() ? tr("Benenne Paar um %1…")
                                                            : tr("Benenne Spieler um %1…")).arg(
                                                                 changeMarkerStartMessage));

    if (! m_db->renamePlayers(originalName, name, markerAfterEditId)) {
        return;
    }

    Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
        ? tr("Paar „%1“ umbenannt zu „%2“%3")
        : tr("Spieler „%1“ umbenannt zu „%2“%3")).arg(
             originalName, name, changeMarkerFinishedMessage));

    const auto originalVisibleRow = m_listView->visibleRow(m_listModel->rowForName(originalName));
    updatePlayersList(name, originalVisibleRow);
    checkUpdateDrawOverviewPage(true, name);

    if (m_isServer) {
        m_server->broadcastRenamePlayers(originalName, name,
                                         markerAfterEditId, markerAfterEditName);
    }

    Q_EMIT nameEdited();
    QApplication::restoreOverrideCursor();
}

void RegistrationPage::updateContextMenu()
{
    const auto &playersData = m_listModel->playersData();
    const auto &selectedData = playersData.at(m_listView->selectedRow());
    const int singleMarker = m_markersWidget->singleMarker();
    const bool isSinglePlayer = selectedData.marker == singleMarker;
    m_assignMenu->menuAction()->setVisible(isSinglePlayer);
    m_contextMenu->setTitle((m_tournamentSettings->isPairMode() && ! isSinglePlayer
                                ? tr("Paar „%1“") : tr("Spieler „%1“")).arg(selectedData.name));
    if (isSinglePlayer) {
        m_assignMenu->clear();
        for (const auto &data : playersData) {
            if (data.marker == singleMarker &&  data.name != selectedData.name) {
                auto *action = m_assignMenu->addAction(data.name);
                connect(action, &QAction::triggered,
                        this, std::bind(&RegistrationPage::assignPlayer, this, data.name));
            }
        }
        if (m_assignMenu->actions().count() == 0) {
            auto *action = m_assignMenu->addAction(tr("keiner sonst angemeldet"));
            action->setEnabled(false);
        }

        m_editAction->setText(tr("Name ändern"));
        m_deleteAction->setText(tr("Spieler löschen"));

    } else if (m_tournamentSettings->isPairMode()) {
        m_editAction->setText(tr("Name(n) ändern"));
        m_deleteAction->setText(tr("Paar löschen"));
    }

    m_removeMarkerAction->setEnabled(selectedData.marker != 0);
    for (auto *action : m_setMarkerMenu->actions()) {
        action->setEnabled(action->data().toInt() != selectedData.marker);
    }

    m_setDrawAction->setVisible(selectedData.assignment.table == 0);
    m_deleteDrawAction->setVisible(selectedData.assignment.table != 0);

    m_disqualificationMenu->setEnabled(selectedData.disqualified || m_db->tournamentStarted());
    m_disqualifyAction->setEnabled(! selectedData.disqualified);
    m_unDisqualifyAction->setEnabled(selectedData.disqualified);
}

void RegistrationPage::showContextMenu(QPoint point)
{
    if (m_listModel->rowCount() == 0 || ! m_db->isChangeable()) {
        return;
    }

    if (m_listView->horizontalHeader()->isVisible()) {
        point.setY(point.y() + m_listView->horizontalHeader()->height());
    }

    if (m_listView->selectedName().isEmpty()) {
        m_listMenu->exec(m_listView->mapToGlobal(point));
    } else {
        updateContextMenu();
        m_contextMenu->exec(m_listView->mapToGlobal(point));
    }
}

void RegistrationPage::updateVisibility()
{
    // Get the row for the currently selected name. If no name is selected
    // and/or the selection isn't present anymore, we will get -1.
    const auto selectedRow = m_listModel->rowForName(m_listView->selectedName());

    if (! m_listModel->searchTerm().isEmpty()) {
        int matches = 0;
        int hiddenMatches = 0;
        for (int row = 0; row < m_listModel->rowCount(); row++) {
            if (! m_listModel->isSearchHidden(row)) {
                matches++;
                if (m_listModel->isMarkerHidden(row)) {
                    hiddenMatches++;
                }
            }
        }

        m_searchWidget->updateMatches(matches, hiddenMatches);
        m_listView->resizeColumnsToContents();
        m_isSearchLabel->setText(tr("(Suchergebnis für „%1“)").arg(
                                    m_listModel->searchTerm().simplified()));
        m_isSearchLabel->show();

        if (m_searchAsYouType->isChecked()) {
            m_searchAsYouTypeResult->setText(m_searchWidget->matchesText());
        } else {
            m_searchAsYouTypeResult->clear();
        }

    } else {
        m_isSearchLabel->hide();
        m_searchAsYouTypeResult->clear();
    }

    int hidden = 0;
    m_actuallyHiddenRow = -1;
    for (int row = 0; row < m_listModel->rowCount(); row++) {
        auto hide = m_listModel->isSearchHidden(row) || m_listModel->isMarkerHidden(row);
        if (hide && row == selectedRow) {
            hide = false;
            m_actuallyHiddenRow = row;
        }
        m_listView->setRowHidden(row, hide);
        hidden += hide;
    }

    if (m_actuallyHiddenRow != -1) {
        m_visibilityDeviation->setText(tr(
            "Die Anmeldung „%1“ wäre jetzt eigentlich ausgeblendet! "
            "<a href=\"#updateVisibility\">ausblenden</a>").arg(
            m_listView->selectedName()));
        m_visibilityDeviation->show();
    } else {
        m_visibilityDeviation->hide();
    }

    if (hidden > 0) {
        m_playersBox->setTitle(m_tournamentSettings->isPairMode()
                                   ? tr("Angemeldete Paare (%n ausgeblendet)", "", hidden)
                                   : tr("Angemeldete Spieler (%n ausgeblendet)", "", hidden));
    } else {
        m_playersBox->setTitle(m_tournamentSettings->isPairMode() ? tr("Angemeldete Paare")
                                                                  : tr("Angemeldete Spieler"));
    }

    m_listView->resizeColumnsToContents();
}

void RegistrationPage::updateSearch(const QString &searchText)
{
    m_listModel->updateSearchHidden(searchText,
        m_searchAsYouType->isChecked() ? m_registerPhoneticSearch->isChecked()
                                       : m_searchWidget->phoneticSearch());
    updateVisibility();
}

void RegistrationPage::resetSearch()
{
    m_searchWidget->clear();
}

void RegistrationPage::updateMarker(const QString &markerName, int markerId)
{
    const auto name = m_listView->selectedName();

    if (m_isClient) {
        m_client->requestUpdateMarker(name, markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (markerId != 0) {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Markiere Paar „%1“ als „%2“ …")
                             : tr("Markiere Spieler „%1“ als „%2“ …")).arg(name, markerName));
    } else {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Entferne Markierung von Paar „%1“ …")
                             : tr("Entferne Markierung von Spieler „%1“ …")).arg(name));
    }

    if (! m_db->updateMarker(name, markerId)) {
        return;
    }

    if (markerId != 0) {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Paar „%1“ als „%2“ markiert")
                             : tr("Spieler „%1“ als „%2“ markiert")).arg(name, markerName));
    } else {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Markierung von Paar „%1“ entfernt")
                             : tr("Markierung von Spieler „%1“ entfernt")).arg(name));
    }

    const auto originalVisibleRow = m_listView->visibleRow(m_listModel->rowForName(name));
    updatePlayersList();
    m_listView->selectName(name, originalVisibleRow);

    if (m_isServer) {
        m_server->broadcastUpdateMarker(name, markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::listUpdateMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestListUpdateMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    if (markerId != 0) {
        Q_EMIT statusUpdate(tr("Markiere alle %1 als „%2“").arg(
                               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                               markerName));
    } else {
        Q_EMIT statusUpdate(tr("Entferne alle Markierungen …"));
    }

    if (! m_db->listUpdateMarker(markerId)) {
        return;
    }

    if (markerId != 0) {
        Q_EMIT statusUpdate(tr("Alle %1 als „%2“ markiert").arg(
                               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"),
                               markerName));
    } else {
        Q_EMIT statusUpdate(tr("Alle Markierungen entfernt"));
    }
    updatePlayersList();

    if (m_isServer) {
        m_server->broadcastListUpdateMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::listRemoveMarker(const QString &markerName, int markerId)
{
    if (m_isClient) {
        m_client->requestListRemoveMarker(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(tr("Entferne alle „%1“-Markierungen …").arg(markerName));

    if (! m_db->listRemoveMarker(markerId)) {
        return;
    }

    Q_EMIT statusUpdate(tr("Alle „%1“-Markierungen entfernt").arg(markerName));
    updatePlayersList();

    if (m_isServer) {
        m_server->broadcastListRemoveMarker(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::deletePlayers()
{
    const auto name = m_listView->selectedName();

    if (QMessageBox::question(this,
            m_tournamentSettings->isPairMode() ? tr("Paar löschen") : tr("Spieler löschen"),
            (m_tournamentSettings->isPairMode()
                ? tr("Soll Paar „%1“ wirklich gelöscht werden?")
                : tr("Soll Spieler „%1“ wirklich gelöscht werden?")).arg(name),
            QMessageBox::Yes | QMessageBox::Cancel,
            QMessageBox::Cancel) == QMessageBox::Cancel) {
        return;
    }

    if (m_isClient || m_isServer) {
        if (! m_listModel->hasExactMatch(name)) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                   "gelöscht oder umbenannt. Das Löschen ist nicht mehr möglich und wird "
                   "abgebrochen.").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                   name));
            return;
        }
    }

    m_listView->deselectRow();

    if (m_isClient) {
        m_client->requestDeletePlayers(name);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                         ? tr("Lösche Paar „%1“ …") : tr("Lösche Spieler „%1“ …")).arg(name));

    if (! m_db->deletePlayers(name)) {
        return;
    }

    Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                         ? tr("Paar „%1“ gelöscht") : tr("Spieler „%1“ gelöscht")).arg(name));

    updatePlayersList();
    checkUpdateDrawOverviewPage(true, name, true);

    if (m_isServer) {
        m_server->broadcastDeletePlayers(name);
    }

    QApplication::restoreOverrideCursor();
}

void RegistrationPage::cacheChecksum()
{
    if (m_isServer || m_isClient) {
        m_cachedChecksum = m_checksum;
    }
}

bool RegistrationPage::checkCachedChecksum(const QString &title)
{
    if (! m_isClient && ! m_isServer) {
        return true;
    }

    if (m_checksum != m_cachedChecksum) {
        QMessageBox::warning(this, title,
            tr("<p><b>Kollidierende Netzwerkänderung</b></p>"
               "<p>Die Anmeldungsliste wurde zwischenzeitlich von einem anderen Netzwerkteilnehmer "
               "verändert. Die Aktion kann nicht mehr durchgeführt werden und wird abgebrochen."
               "</p>"));
        return false;
    }

    return true;
}

bool RegistrationPage::confirmListDeleteAction(const QString &title, const QString &text)
{
    m_backupFile.clear();

    QMessageBox messageBox(QMessageBox::Warning, title, text,
                           QMessageBox::Yes | QMessageBox::Cancel, this);
    messageBox.setDefaultButton(QMessageBox::Cancel);
    QCheckBox *createBackup = new QCheckBox(tr("Vorher ein Backup der Datenbank anlegen"));
    createBackup->setChecked(true);
    messageBox.setCheckBox(createBackup);

    if (messageBox.exec() == QMessageBox::Cancel) {
        return false;
    }

    if (createBackup->isChecked()) {
        const QString openedDb = m_db->dbFile();
        BackupEngine backupEngine(openedDb);
        if (! m_db->saveSettings()) {
            return false;
        }
        m_db->closeDatabaseConnection();
        bool success = backupEngine.createBackup();
        m_db->openDatabase(openedDb);

        if (! success) {
            QMessageBox::critical(this, title,
                tr("<p><b>Das Anlegen eines Backups ist fehlgeschlagen!</b></p>"
                   "<p>Die Datei konnte nicht angelegt werden. Bitte die Zugriffsrechte von „%1“ "
                   "bzw. dem Verzeichnis „%2“) überprüfen!</p>").arg(
                   backupEngine.backupFile(), QDir::toNativeSeparators(backupEngine.path())));
            return false;
        }

        m_backupFile = backupEngine.backupFile();
    }

    return true;
}

void RegistrationPage::checkDisplayBackup(const QString &title)
{
    if (! m_backupFile.isEmpty()) {
        QMessageBox::information(this, title,
            tr("Es wurde erfolgreich ein Backup der Datenbank mit dem Dateinamen „%1“ "
               "erstellt.").arg(m_backupFile));
    }
}

void RegistrationPage::deleteMarked(const QString &markerName, int markerId)
{
    cacheChecksum();
    const QString title = tr("Markierte %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle als „%1“ markierten %2 gelöscht werden?").arg(markerName,
               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteMarkedPlayers(markerName, markerId);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(tr("Lösche alle als „%1“ markierten %2 …").arg(
                           markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                          : tr("Spieler")));
    if (! m_db->deleteMarkedPlayers(markerId)) {
        return;
    }

    Q_EMIT statusUpdate(tr("Alle als „%1“ markierten %2 gelöscht").arg(
                           markerName, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                          : tr("Spieler")));

    updatePlayersList();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteMarkedPlayers(markerName, markerId);
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::deleteUnmarked()
{
    cacheChecksum();
    const QString title = tr("Unmarkierte %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
            tr("Sollen wirklich alle unmarkierten %1 gelöscht werden?").arg(
               m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))
        || ! checkCachedChecksum(title)) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteUnmarkedPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(tr("Lösche alle unmarkierten %1 …").arg(
                           m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteUnmarkedPlayers()) {
        return;
    }

    Q_EMIT statusUpdate(tr("Alle unmarkierten %1 gelöscht").arg(
                           m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    updatePlayersList();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteUnmarkedPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::deleteAll()
{
    const QString title = tr("Alle %1 löschen").arg(
                             m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler"));

    if (! confirmListDeleteAction(title,
              tr("Sollen wirklich alle %1 gelöscht werden?").arg(
                 m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")))) {

        return;
    }

    if (m_isClient) {
        m_client->requestDeleteAllPlayers();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(tr("Lösche alle %1 …").arg(
                           m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    if (! m_db->deleteAllPlayers()) {
        return;
    }

    Q_EMIT statusUpdate(tr("Alle %1 gelöscht").arg(
                           m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));

    updatePlayersList();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteAllPlayers();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::setColumnVisible(rc::Column column)
{
    const bool state = m_showColumn.value(column)->isChecked();

    // If we're about to hide the draw column, potentially ask if the user really wants this

    if (column == rc::Draw
        && ! state
        && ! m_tournamentSettings->loadingDb()
        && ! m_dismissableMessage->dismissed(DismissableMessage::AskHideDrawColumn)) {

        if (m_listModel->assignmentPresent()
            && m_dismissableMessage->question(DismissableMessage::AskHideDrawColumn, this,
               tr("Spalte „Auslosung“ ausblenden"),
               tr("Es wurden bereits Auslosungen eingegeben. Soll die Spalte „Auslosung“ wirklich "
                  "ausgeblendet werden?")) == QMessageBox::No) {

            // Re-check the respective context menu entry
            auto *action = m_showColumn[rc::Draw];
            action->blockSignals(true);
            action->setChecked(true);
            action->blockSignals(false);

            // Also check the respective main menu action, in case this came from there
            Q_EMIT columnVisibilityChanged(rc::Draw, true);

            return;
        }
    }

    // Inform the MainWindow so that the menu can be updated
    Q_EMIT columnVisibilityChanged(column, state);

    m_listView->setColumnHidden(rc::columnIndex.value(column), state == Qt::Unchecked);
    m_listView->resizeColumnsToContents();

    int lastVisibleColumn = 0;
    for (int i = 0; i < rc::columnIndex.count(); i++) {
        if (i == rc::columnIndex.value(column) && state == Qt::Unchecked) {
            continue;
        }
        if (! m_listView->isColumnHidden(i) || (i == rc::columnIndex.value(column) && state)) {
            lastVisibleColumn = i;
        }
    }
    m_listDelegate->setLastVisibleColumn(lastVisibleColumn);
    m_listView->setColumnHidden(rc::columnIndex.value(column), state == Qt::Unchecked);
    m_listView->resizeColumnsToContents();

    if (column == rc::Column::Draw && m_db->isOpen()) {
        // We need to call this to update the status
        updatePlayersList();
    }

    // Inform the database
    QVector<rc::Column> columns;
    for (const rc::Column column : rc::allColumns) {
        if (m_showColumn.value(column)->isChecked()) {
            columns.append(column);
        }
    }
    m_tournamentSettings->setRegistrationColumns(columns);

    // If the draw column is shown, ask if the draw overview page should also be shown
    if (column == rc::Column::Draw && state
        && ! m_dismissableMessage->dismissed(DismissableMessage::AskShowDrawOverview)
        && ! m_tournamentSettings->loadingDb()
        && m_drawOverviewPage == nullptr
        && m_dismissableMessage->question(DismissableMessage::AskShowDrawOverview, this,
               tr("Spalte „Auslosung“ anzeigen"),
               tr("Soll die Auslosungs-Übersichts-Seite ebenfalls angezeigt werden?"))
           == QMessageBox::Yes) {

        Q_EMIT showDrawOverviewPage();
        QTimer::singleShot(0, qobject_cast<QWidget *>(parent()), &QWidget::raise);
    }
}

void RegistrationPage::updateChecksum()
{
    ChecksumHelper checksumHelper;

    // Add the players data
    for (const auto &data : m_listModel->playersData()) {
        checksumHelper.addData(data);
    }

    // Add the marker data
    m_markersWidget->addChecksumData(checksumHelper);

    // Add the tournament state
    checksumHelper.addData(m_db->networkTournamentStarted());

    m_checksum = checksumHelper.checksum();
}

const QString &RegistrationPage::checksum() const
{
    return m_checksum;
}

void RegistrationPage::serverStarted(Server *server)
{
    updateChecksum();
    m_server = server;
    m_isServer = true;
    m_markersWidget->serverStarted(server);
}

void RegistrationPage::serverStopped()
{
    m_isServer = false;
    m_markersWidget->serverStopped();
    markersChanged();
    m_checksum.clear();
}

void RegistrationPage::hideContextMenus()
{
    const auto contextMenus = findChildren<QMenu *>();
    for (QMenu *menu : contextMenus) {
        menu->hide();
    }
}

void RegistrationPage::requestRejected()
{
    if (m_requestingRegistration) {
        m_requestingRegistration = false;
    }
}

void RegistrationPage::processNetworkChange(int clientId)
{
    // Damn. This is one nasty function. But screw it. We need it :-P

    // Restore all override cursors. None should be set at this point, but whatever.
    while (QApplication::overrideCursor() != nullptr) {
        QApplication::restoreOverrideCursor();
    }

    if (m_isClient && clientId == m_client->id()) {
        // We're processing our own request. No edit can be going on, no context menu can be
        // visible, and no popup can be shown in this case. We can simply display the change.

        if (m_requestingRegistration) {
            m_requestingRegistration = false;
            m_playersName->clear();
        }

        if (m_focusPlayersName) {
            m_focusPlayersName = false;
            m_playersName->setFocus();
        }

        const auto originalVisibleRow = m_listView->selectedVisibleRow();
        updatePlayersList();

        if (! m_scrollToTarget.isEmpty()) {
            const auto row = m_listModel->rowForName(m_scrollToTarget);
            if (row == -1) {
                itemError(Q_FUNC_INFO, m_scrollToTarget);
                return;
            }
            m_listView->selectRow(row, originalVisibleRow);
        }

        return;
    }

    // We're processing a change from another client or the server. It's possibly a bit more
    // complicated in this case, as maybe, we're just about to do something.

    // Cache what we have

    const auto selectedName = m_listView->selectedName();
    const auto selectedRow = m_listView->selectedRow();
    const auto originalVisibleRow = m_listView->selectedVisibleRow();
    const auto originalDraw = selectedRow != -1 ? m_listModel->data(selectedRow).assignment
                                                : Draw::EmptySeat;

    const auto editorOpen = m_listView->isEditing();
    if (editorOpen) {
        m_listView->saveEditorState();
    }

    // Update the list
    updatePlayersList();

    // In each case, we try to re-select the previously selected name and try to keep it's position

    const auto targetRow = m_listModel->rowForName(selectedName);
    const auto selectedNamePresent = ! selectedName.isEmpty()
                                     && targetRow != -1
                                     && ! m_listView->isRowHidden(targetRow);

    if (selectedNamePresent) {
        // Re-select the previously selected name
        m_listView->selectRow(targetRow, originalVisibleRow);
        const auto processedVisibleRow = m_listView->selectedVisibleRow();

        // Try to scroll so that it's at the same place on screen if the position is different
        if (processedVisibleRow != originalVisibleRow) {
            m_listView->verticalScrollBar()->setValue(
                m_listView->verticalScrollBar()->value()
                - originalVisibleRow + processedVisibleRow);
        }
    }

    // Check if we had the whole list menu and if so, if we can still display it

    if (m_listMenu->isVisible()) {
        // The only thing that would stop us is if somebody deleted all registrations
        if (m_listModel->rowCount() == 0) {
            QTimer::singleShot(0, [this]
            {
                // Showing this warning will also close all context menus
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("Es wurden von einem anderen Netzwerkteilnehmer alle Anmeldungen "
                       "gelöscht. Ein Bearbeiten der ganzen Liste ist nicht mehr möglich."));
            });
            return;
        }

        if (m_contextMenu->isVisible() && ! selectedNamePresent) {
            // If the selected name isn't present anymore, we still can keep showing the whole list
            // menu. Hiding the context menu will also hide the whole list menu, so we have to show
            // it again afterwards
            QTimer::singleShot(0, [this]
            {
                const auto pos = m_listMenu->geometry().topLeft();
                m_contextMenu->hide();
                m_listMenu->exec(pos);
            });
        }

        return;
    }

    // If the list menu wasn't visible, and no name was selected before the change, no action was
    // going on. In this case, we're done here.
    if (selectedName.isEmpty()) {
        return;
    }

    // For everything else, we need the previously selected name to be still present.

    if (! selectedNamePresent) {
        // The selected name is not present anymore. We have to abort here.

        // Remove the selection
        m_listView->deselectRow();

        // Close a possibly shown "fuzzy match" warning
        if (m_fuzzyMatchWarning != nullptr) {
            m_fuzzyMatchWarning->reject();
        }

        // Inform the user
        QTimer::singleShot(0, [this, selectedName]
        {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("<p>%1 „%2“ wurde durch eine Änderung eines anderen Netzwerkteilnehmers "
                   "entweder umbenannt oder gelöscht.</p>"
                   "<p>Die Auswahl bzw. die laufende Bearbeitung kann nicht wiederhergestellt "
                   "werden.</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("Das gerade bearbeitete Paar")
                                                      : tr("Der gerade bearbeitete Spieler"),
                   selectedName));
        });

        return;
    }

    // If we were editing before the change, we restore the editor's state. In this case, no
    // context menu or popup could have been shown and we're done here.
    if (editorOpen) {
        m_listView->restoreEditorState();
        return;
    }

    // If the context menu is visible, we have to update it. Markers could have changed etc.
    if (m_contextMenu->isVisible()) {
        updateContextMenu();
        return;
    }

    // Update the markers popup if it's visible
    if (m_markersPopup->isVisible()) {
        popupRequested(rc::Mark, m_markersPopup->geometry().topLeft());
        return;
    }

    // Update the draw popup if it's visible
    if (m_drawPopup->isVisible()) {
        const auto &data = m_listModel->data(m_listView->selectedRow());

        // The entry could have been marked as absent
        if (data.marker == m_tournamentSettings->marker(Markers::Absent)) {
            QTimer::singleShot(0, [this, selectedName]
            {
                m_drawPopup->reject();
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("<p>%1 „%2“ wurde durch eine Änderung eines anderen Netzwerkteilnehmers "
                       "als „abwesend“ markiert.</p>"
                       "<p>Das Eingeben einer Auslosung ist nicht mehr möglich und wurde "
                       "abgebrochen.</p>").arg(
                       m_tournamentSettings->isPairMode() ? tr("Das gerade bearbeitete Paar")
                                                          : tr("Der gerade bearbeitete Spieler"),
                       selectedName));
            });
            return;
        }

        // If a draw has been entered by another network user for the processed name,
        // we have to re-show the dialog, with everything else being already processed
        if (m_listModel->data(m_listView->selectedRow()).assignment != originalDraw) {
            m_drawPopup->reject();
            QTimer::singleShot(0, [this, selectedName]
            {
                QMessageBox::warning(this,
                    tr("Kollidierende Netzwerkänderung"),
                    tr("<p>Für %1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer "
                       "eine Auslosung eingegeben.</p>"
                       "<p>Sofern das wirklich gewollt ist, dann bitte die Änderung der Auslosung "
                       "erneut angfragen.</p>").arg(
                       m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("den Spieler"),
                       selectedName));
            });
        }

        return;
    }
}

void RegistrationPage::clientConnected(Client *client)
{
    m_isClient = true;
    m_client = client;
    m_markersWidget->clientConnected(client);
}

void RegistrationPage::clientDisconnected()
{
    m_isClient = false;
    m_markersWidget->clientDisconnected();
    markersChanged();
    m_checksum.clear();
    m_requestingRegistration = false;
}

void RegistrationPage::setProcessingRequest(bool processing, const QString &scrollToTarget)
{
    m_newPlayersBox->setEnabled(! processing);
    m_playersBox->setEnabled(! processing);
    if (processing) {
        m_scrollToTarget = scrollToTarget;
        QApplication::setOverrideCursor(Qt::WaitCursor);
    }
}

MarkersWidget *RegistrationPage::markersWidget() const
{
    return m_markersWidget;
}

void RegistrationPage::finishRegistration()
{
    if (QMessageBox::warning(this,
        tr("Anmeldung beenden"),
        tr("<p>Wenn die Anmeldung beendet ist, kann (mindestens während das Netzwerk läuft) kein "
           "Netzwerkteilnehmer mehr neue %1 anmelden oder löschen!</p>"
           "<p>Soll die Anmeldung wirklich beendet werden?</p>").arg(
           m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")),
        QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel) == QMessageBox::Cancel) {

        return;
    }

    if (m_isClient) {
        m_client->requestFinishRegistration();
        return;
    }

    // Otherwise, we're the server
    networkFinishRegistration();
    m_server->broadcastFinishRegistration();
}

void RegistrationPage::networkFinishRegistration(int clientId)
{
    QString networkChange;
    if ((m_isClient && clientId != m_client->id()) || (m_isServer && clientId != 0)) {
        networkChange = tr(" von einem anderen Netzwerkteilnehmer");
        if (m_contextMenu->isVisible()) {
            hideContextMenus();
        }
        if (m_listView->isEditing()) {
            m_listView->abortEdit();
        }
    }

    m_db->setNetworkTournamentStarted(true);
    reload();
    Q_EMIT networkTournamentStarted(true);

    QTimer::singleShot(0, [this, networkChange]
    {
        QMessageBox::information(this,
            tr("Anmeldung beendet"),
            tr("<p>Die Anmeldung wurde%1 beendet. Es können ab jetzt keine neuen %2 mehr "
               "angemeldet oder bestehende Anmeldungen gelöscht werden.</p>"
               "<p>Sollte das trotzdem noch nötig sein, dann bitte den Server stoppen, neu "
               "starten und alle Clients neu verbinden. Sofern noch keine Ergebnisse eingegeben "
               "wurden, ist danach das komplette Bearbeiten der Anmeldungsliste wieder möglich."
               "</p>").arg(networkChange, m_tournamentSettings->isPairMode() ? tr("Paare")
                                                                             : tr("Spieler")));
    });
}

void RegistrationPage::popupRequested(rc::Column column, QPoint pos)
{
    if (! m_db->isChangeable()) {
        return;
    }

    if (column == rc::Mark) {
        m_markersPopup->show(m_listModel->data(m_listView->selectedRow()).marker,
                             pos == QPoint(-1, -1) ? QCursor::pos() : pos);

    } else if (column == rc::Draw) {
        if (! m_showColumn.value(rc::Draw)->isChecked()) {
            askShowDrawColumn();
        }

        const auto &data = m_listModel->data(m_listView->selectedRow());

        if (data.marker == m_tournamentSettings->marker(Markers::Absent)
            && data.assignment.isEmpty()) {

            QMessageBox::warning(this,
                tr("Auslosung eingeben"),
                tr("<p><b>Auslosung eingeben</b></p>")
                + (m_tournamentSettings->isPairMode()
                       ? tr("<p>Das Paar „%1“ ist momentan als abwesend markiert. Für ein "
                            "abwesendes Paar kann keine Auslosung eingegeben werden."
                            "</p>").arg(m_listView->selectedName())
                       : tr("<p>Der Spieler „%1“ ist momentan als abwesend markiert. Für einen "
                            "abwesenden Spieler kann keine Auslosung eingegeben "
                            "werden.").arg(m_listView->selectedName())));
            return;
        }

        m_drawPopup->show(data.assignment, m_db->allRoundsFinished(),
                          data.marker == m_markersWidget->singleMarker(),
                          pos == QPoint(-1, -1) ? QCursor::pos() : pos);
    }
}

void RegistrationPage::askShowDrawColumn()
{
    if (m_dismissableMessage->dismissed(DismissableMessage::AskShowDrawColumn)
        || m_showColumn.value(rc::Column::Draw)->isChecked()) {

        return;
    }

    if (m_dismissableMessage->question(DismissableMessage::AskShowDrawColumn, this,
            tr("Auslosung speichern"),
            tr("Momentan ist die Spalte „Auslosung“ auf der Anmeldeseite ausgeblendet. Soll sie "
               "angezeigt werden?")) == QMessageBox::Yes) {

        m_showColumn[rc::Column::Draw]->setChecked(true);
    }
}

void RegistrationPage::updateDraw(const Draw::Seat &seat)
{
    // Check for a suspiciously high table number
    if (seat.table >= DefaultValues::drawTableWarning
        && QMessageBox::question(this,
               tr("Auslosung speichern"),
               tr("Soll wirklich eine Auslosung für Tisch Nr. %1 eingegeben werden (vielleicht "
                  "war es ein Tippfehler)?").arg(QString::number(seat.table)),
               QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
        == QMessageBox::Cancel) {

        return;
    }

    const auto name = m_listView->selectedName();

    if (m_isClient || m_isServer) {
        // If we're on a network, we have to check if somebody else did
        // an assignment for the registration we're working on meanwhile

        if (! seat.isEmpty() // We want to delete an assignment otherwise
            && ! m_listModel->data(m_listView->selectedRow()).assignment.isEmpty()) {

            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("<p>Für %1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer eine "
                   "Auslosung eingegeben.</p>"
                   "<p>Sofern das wirklich gewollt ist, dann bitte die Änderung der Auslosung "
                   "erneut angfragen.</p>").arg(
                   m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("den Spieler"),
                   name));

            return;
        }

        // ... and we also have to check if the very seat we're about to assign is stll available
        if (m_drawPopup->isSeatTaken(seat)) {
            QMessageBox::warning(this,
                tr("Kollidierende Netzwerkänderung"),
                tr("<p>Die Auslosung „Tisch %1 Paar %2“ wurde mittlerweile durch einen anderen "
                   "Netzwerkteilnehmer vergeben und kann nicht mehr zugewiesen werden.</p>"
                   "<p>Bitte die Auslosung neu eingeben.</p>").arg(
                   QString::number(seat.table), QString::number(seat.pair)));
            return;
        }
    }

    const int marker = seat.isEmpty() ? -1 : m_markersWidget->drawnMarker();

    if (m_isClient) {
        m_client->requestUpdateDraw(name, seat, marker);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(seat.isEmpty() ? tr("Lösche Auslosung …") : tr("Speichere Auslosung …"));

    if (! m_db->updateDraw(name, seat, marker)) {
        return;
    }

    if (seat.isEmpty()) {
        Q_EMIT statusUpdate(tr("Auslosung für %1 „%2“ gelöscht").arg(
                               m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                               name));
    } else {
        Q_EMIT statusUpdate(tr("Auslosung für %1 „%2“ gespeichert (Tisch %3 Paar %4)").arg(
                               m_tournamentSettings->isPairMode() ? tr("Paar") : tr("Spieler"),
                               name, QString::number(seat.table), QString::number(seat.pair)));
    }

    const auto originalVisibleRow = m_listView->visibleRow(m_listModel->rowForName(name));
    updatePlayersList();
    checkUpdateDrawOverviewPage(true, name, seat.isEmpty());
    m_listView->selectName(name, originalVisibleRow);

    if (m_isServer) {
        m_server->broadcastUpdateDraw(name, seat, marker);
    }

    Q_EMIT drawChanged();
    QApplication::restoreOverrideCursor();
}

void RegistrationPage::assignPlayer(const QString &name)
{
    const auto selectedName = m_listView->selectedName();
    const auto pairName = selectedName + m_settings->namesSeparator() + name;

    if (m_listModel->hasCaseInsensitiveMatch(pairName)) {
        QMessageBox::warning(this, tr("Allein gekommene Spieler zuordnen"),
            tr("<p><b>Zuordnung nicht möglich</b></p>"
               "<p>Der Spieler „%1“ kann dem Spieler „%2“ nicht zugeordnet werden, da der Paarname "
               "„%3“ bereits vergeben ist.</p>"
               "<p>Bitte mindestestens einen Namen ändern, damit ein eindeutiger Paarname erstellt "
               "werden kann!</p>").arg(
               name, selectedName, pairName));
        return;
    }

    if (m_isClient) {
        m_client->requestAssemblePair(selectedName, name, m_settings->namesSeparator(),
                                      m_markersWidget->assignedMarker());
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(tr("Erstelle Paar aus Einzelspielern …"));

    if (! m_db->assemblePair(selectedName, name, m_settings->namesSeparator(),
                             m_markersWidget->assignedMarker())) {
        return;
    }

    Q_EMIT statusUpdate(tr("Paar „%1“ erstellt").arg(pairName));

    updatePlayersList();
    if (! m_listModel->hasExactMatch(pairName)) {
        itemError(Q_FUNC_INFO, pairName);
        return;
    }
    checkUpdateDrawOverviewPage(true, pairName);

    m_listView->selectName(pairName);

    if (m_isServer) {
        m_server->broadcastAssemblePair(selectedName, name, m_settings->namesSeparator(),
                                        m_markersWidget->assignedMarker());
    }

    QApplication::restoreOverrideCursor();
}

bool RegistrationPage::absentsPresent() const
{
    return m_absentsPresent;
}

bool RegistrationPage::singlesPresent() const
{
    return m_singlesPresent;
}

void RegistrationPage::deleteAllDraws()
{
    const QString title = tr("Alle Auslosungen löschen");
    if (! confirmListDeleteAction(title, tr("Sollen wirklich alle bisher eingegebenen Auslosungen "
                                            "gelöscht werden?"))) {
        return;
    }

    if (m_isClient) {
        m_client->requestDeleteAllDraws();
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(tr("Lösche alle Auslosungen …"));

    if (! m_db->deleteAllDraws()) {
        return;
    }

    Q_EMIT drawChanged();
    Q_EMIT statusUpdate(tr("Alle Auslosungen gelöscht"));
    updatePlayersList();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastDeleteAllDraws();
    }

    QApplication::restoreOverrideCursor();
    checkDisplayBackup(title);
}

void RegistrationPage::processStatusLinkClick(const QString &anchor)
{
    if (anchor == QLatin1String("#updateVisibility")) {
        if (m_listView->selectedRow() == m_actuallyHiddenRow) {
            m_listView->deselectRow();
        }
        updateVisibility();
        return;
    }

    if (anchor == QLatin1String("#drawOverview")) {
        Q_EMIT showDrawOverviewPage();
        return;
    }

    if (anchor == QLatin1String("#adjustDraw")) {
        Q_EMIT showAdjustDrawDialog();
        return;
    }

    QString text;

    if (anchor == QLatin1String("#unfinished")) {
        text = tr("<p><b>Unvollständige Auslosung</b></p>");

        QString list;
        int missingDraws = 0;
        for (const auto &data : m_listModel->playersData()) {
            if (data.assignment.isEmpty()) {
                missingDraws++;
                if (missingDraws <= 10) {
                    list.append(QStringLiteral("<li>%1</li>").arg(data.name));
                }
            }
        }

        text.append(m_tournamentSettings->isPairMode()
            ?  tr("<p>Das/die folgende(n) Paar(e) hat/haben noch keine Auslosung:</p><ul>%1</ul>",
                  "", missingDraws).arg(list)
            :  tr("<p>Der/die folgende(n) Spieler hat/haben noch keine Auslosung:</p><ul>%1</ul>",
                  "", missingDraws).arg(list));

        if (missingDraws > 10) {
            text.append(m_tournamentSettings->isPairMode()
                ? tr("<p>und %n weitere(s)</p>", "für Paare", missingDraws - 10)
                : tr("<p>und %n weitere(r)</p>", "für Spieler", missingDraws - 10));
        }

    } else if (anchor == QLatin1String("#incompleteTables")) {
        text = tr("<p><b>Tische mit unvollständiger Auslosung</b></p>");
        const int incompleteCount = m_listModel->incompleteTables().count();
        if (incompleteCount == 1) {
            text.append(tr("<p>Zu Tisch <b>%1</b> wurde bisher nur ein Paar 1 oder ein Paar 2 "
                           "zugelost%2.</p>").arg(
                           QString::number(m_listModel->incompleteTables().at(0)),
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
        } else if (incompleteCount == 2) {
            text.append(tr("<p>Zu den Tischen <b>%1</b> und <b>%2</b> wurden bisher nur ein Paar 1 "
                           "oder ein Paar 2 zugelost%3.</p>").arg(
                           QString::number(m_listModel->incompleteTables().at(0)),
                           QString::number(m_listModel->incompleteTables().at(1)),
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
        } else {
            text.append(tr("<p>Zu folgenden Tischen wurden bisher nur ein Paar 1 oder ein Paar 2 "
                           "zugelost%2:</p><p>").arg(
                           m_tournamentSettings->isPairMode()
                               ? QString() : tr(" (bzw. die Paare sind unvollständig)")));
            for (int i = 0; i < incompleteCount - 1; i++) {
                text.append(tr("<b>%1</b>, ").arg(
                               QString::number(m_listModel->incompleteTables().at(i))));
            }
            text = text.left(text.size() - 2);
            text.append(tr(" und <b>%1</b>.</p>").arg(
                           m_listModel->incompleteTables().at(incompleteCount - 1)));
        }

    } else if (anchor == QLatin1String("#tableGaps")) {
        text = tr("<p><b>Tische, die in der Auslosung fehlen</b></p>");
        const int gapsCount = m_listModel->tableGaps().count();
        if (gapsCount == 1) {
            text.append(tr("<p>Tisch <b>%1</b> ist bisher nicht ausgelost, müsste es aber sein (da "
                           "es sonst Lücken gibt)</p>").arg(
                           QString::number(m_listModel->tableGaps().at(0))));
        } else if (gapsCount == 2) {
            text.append(tr("<p>Die Tische <b>%1</b> und <b>%2</b> sind bisher nicht zugewiesen, "
                           "müssten es aber sein (da es sonst Lücken gibt).</p>").arg(
                           QString::number(m_listModel->tableGaps().at(0)),
                           QString::number(m_listModel->tableGaps().at(1))));
        } else {
            text.append(tr("<p>Folgende Tische sind bisher nicht ausgelost, müssten es aber "
                           "sein (da es sonst Lücken gibt):</p><p>"));
            for (int i = 0; i < gapsCount - 1; i++) {
                text.append(tr("<b>%1</b>, ").arg(QString::number(m_listModel->tableGaps().at(i))));
            }
            text = text.left(text.size() - 2);
            text.append(tr(" und <b>%1</b>.</p>").arg(m_listModel->tableGaps().at(gapsCount - 1)));
        }
    }

    QMessageBox::information(this, tr("Auslosung"), text);
}

void RegistrationPage::setDrawOverviewPage(DrawOverviewPage *drawOverviewPage)
{
    m_drawOverviewPage = drawOverviewPage;
    updatePlayersList();
}

void RegistrationPage::checkUpdateDrawOverviewPage(bool ownChange, const QString &target,
                                                   bool deleted) const
{
    if (m_drawOverviewPage != nullptr) {
        m_drawOverviewPage->updateDraw(ownChange, target, deleted);
    }
}

void RegistrationPage::setColumnsVisible(const QVector<rc::Column> &columns)
{
    // We have to call this via a QTimer call if the database is opened via command line argument
    // to be sure that everything is constructed and set up before we do this
    QTimer::singleShot(0, [this, columns]
    {
        for (const rc::Column column : rc::allColumns) {
            m_showColumn[column]->setChecked(columns.contains(column));
        }
    });
}

void RegistrationPage::showDrawSettings()
{
    auto dialog = DrawSettingsDialog(this);
    dialog.updateSettings(m_tournamentSettings->drawSettings());
    if (dialog.exec()) {
        m_tournamentSettings->setDrawSettings(dialog.settings());
        updatePlayersList();
        if (m_listView->isColumnHidden(rc::columnIndex.value(rc::Column::Draw))) {
            askShowDrawColumn();
        }
    }
}

void RegistrationPage::drawAllSeats()
{
    // First check if we can actually do a draw for all seats and tell the user why not if not

    if (! m_db->playersDividable() || m_singlesPresent || m_absentsPresent) {
        QString whyNot;
        if (! m_db->playersDividable()) {
            whyNot.append(tr("<li>Es ist nicht die passende Anzahl %1 angemeldet</li>").arg(
                m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
        }
        if (m_singlesPresent) {
            whyNot.append(tr("<li>Es sind noch Spieler als „allein da“ markiert</li>"));
        }
        if (m_absentsPresent) {
            whyNot.append(tr("<li>Es noch %1 als „abwesend“ markiert</li>").arg(
                m_tournamentSettings->isPairMode() ? tr("Paare") : tr("Spieler")));
        }

        QMessageBox::warning(this, tr("Alle Plätze auslosen"),
            tr("<p><b>Alle Plätze auslosen</b></p>"
               "<p>Das Auslosen aller Plätze ist derzeit nicht möglich:</p>"
               "<ul>%1</ul>").arg(whyNot));

        return;
    }

    cacheChecksum();

    if (m_listView->isColumnHidden(rc::columnIndex.value(rc::Column::Draw))) {
        askShowDrawColumn();
    }

    const int tablesLimit = m_tournamentSettings->drawSettings().tablesLimit;
    const int neededTables = m_listModel->rowCount() / m_db->necessaryDivisor();
    if (tablesLimit != -1 && neededTables > tablesLimit) {
        // The number of registrations would need more tables than the maximum number of tables
        // that has been defined in the draw settings
        const QString title = tr("Alle Plätze auslosen");
        if (QMessageBox::warning(this, title,
                tr("<p><b>Alle Plätze auslosen</b></p>"
                   "<p>Die Anzahl der Tische, die momentan für eine komplette Auslosung benötigt "
                   "werden (%1), ist höher, als das Tisch-Limit, das in den Einstellungen für die "
                   "Auslosung festgelegt wurde (%2)!</p>"
                   "<p>Sollen wirklich alle Plätze ausgelost werden?</p>").arg(
                   QString::number(neededTables), QString::number(tablesLimit)),
                QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
                == QMessageBox::Cancel
            || ! checkCachedChecksum(title)) {

            return;
        }
    }

    if (m_deleteAllDrawsAction->isEnabled()) {
        // We have existing entered draws
        const QString title = tr("Alle Plätze auslosen");
        if (QMessageBox::warning(this, title,
                tr("<p><b>Alle Plätze auslosen</b></p>"
                   "<p>Es wurden bereits Auslosungen eingegeben. Wenn alle Plätze ausgelost "
                   "werden, dann werden alle bestehenden Auslosungen überschrieben und neu "
                   "vergeben.</p>"
                   "<p>Sollen wirklich alle Plätze ausgelost werden?</p>"),
                QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel)
                == QMessageBox::Cancel
            || ! checkCachedChecksum(title)) {

            return;
        }
    }

    if (m_db->tournamentStarted() && ! m_db->allRoundsFinished()) {
        const QString title = tr("Alle Plätze auslosen");
        if (QMessageBox::warning(this, title,
                tr("<p><b>Alle Plätze auslosen</b></p>"
                   "<p>Es wurden in mindestens einer Runde noch nicht alle Ergebnisse eingegeben. "
                   "Wenn jetzt neu ausgelost wird, dann wird das evtl. dazu führen, dass manche "
                   "Ergebnisse nicht entsprechend der Auslosung eingegeben werden können, weil "
                   "bereits ein Ergebnis für einen anders ausgelosten Tisch eingegeben wurde.</p>"
                   "<p>Soll die Auslosung wirklich neu erstellt werden?</p>"),
                QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No
            || ! checkCachedChecksum(title)) {

            return;
        }
    }

    const auto draw = m_drawPopup->completeDraw(m_db->pairsOrPlayersCount());
    setListDraw(draw);
}

void RegistrationPage::setListDraw(const QVector<Draw::Seat> &draw)
{
    // Here, we need a list of all names
    QVector<QString> playersNames;
    for (const auto &data : m_listModel->playersData()) {
        playersNames.append(data.name);
    }

    if (m_isClient) {
        m_client->requestSetListDraw(playersNames, draw);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);
    Q_EMIT statusUpdate(tr("Setze alle Auslosungen …"));

    if (! m_db->setListDraw(playersNames, draw)) {
        return;
    }

    Q_EMIT drawChanged();
    Q_EMIT statusUpdate(tr("Alle Auslosungen gesetzt"));
    updatePlayersList();
    checkUpdateDrawOverviewPage(true);

    if (m_isServer) {
        m_server->broadcastSetListDraw(playersNames, draw);
    }

    QApplication::restoreOverrideCursor();
}

bool RegistrationPage::registerPhoneticSearchEnabled() const
{
    return m_registerPhoneticSearch->isChecked();
}

void RegistrationPage::adoptDrawAdjustment(const QVector<QString> &names,
                                           const QVector<Draw::Seat> &seats)
{
    QVector<Draw::Seat> newDraw;
    for (int i = 0; i < m_listModel->rowCount(); i++) {
        const auto &data = m_listModel->data(i);
        if (names.contains(data.name)) {
            newDraw.append(seats.at(names.indexOf(data.name)));
        } else {
            newDraw.append(data.assignment);
        }
    }
    setListDraw(newDraw);
}

void RegistrationPage::emitDrawAdjustmentNeeded() const
{
    Q_EMIT drawAdjustmentNeeded(m_drawAdjustmentNeeded);
}

bool RegistrationPage::drawAdjustmentNeededState() const
{
    return m_drawAdjustmentNeeded;
}

void RegistrationPage::adoptSearchText()
{
    const QString text = m_searchWidget->searchText().simplified();
    const auto words = text.split(QStringLiteral(" "));
    QString newText;
    for (const auto &word : words) {
        newText += QStringLiteral("%1%2 ").arg(m_stringTools->toUpper(word.at(0)),
                                               word.mid(1, word.length()));
    }

    m_playersName->setText(newText.simplified());
    m_searchWidget->clear();
    m_playersName->setFocus();
}

void RegistrationPage::showHeaderContextMenu(const QPoint &point)
{
    m_showHideHeaderHint = true;
    m_menuButton->clearFocus();
    m_columnsMenu->setTitleHidden(false);
    m_columnsMenu->exec(m_listView->mapToGlobal(point));
    m_columnsMenu->setTitleHidden(true);
    m_showHideHeaderHint = false;
}

void RegistrationPage::setHideHeader(bool state)
{
    // Inform the MainWindow so that the main menu can be updated
    Q_EMIT headerVisibilityChanged(state);

    if (m_showHideHeaderHint) {
        QMessageBox::information(this, tr("Kopfzeile ausblenden"),
            tr("<p><b>Kopfzeile ausblenden</b></p>"
               "<p>Die Kopfzeile der Anmeldungsliste kann über den „Einstellungen“-Knopf oder über "
               "das Hauptmenü („Ansicht“ → „Anmeldeseite“) wieder eingeblendet werden.</p>"));
    }

    m_listView->horizontalHeader()->setVisible(! state);

    m_tournamentSettings->setRegistrationListHeaderHidden(state);
}

void RegistrationPage::itemError(const char *funcInfoText, const QString &text)
{
    qCCritical(MuckturnierLog) << "Error:" << funcInfoText << "\n   Item not found:" << text;
    QTimer::singleShot(0, [this, funcInfoText, text]
    {
        QMessageBox::critical(this, tr("Unerwarteter Fehler"),
            tr("<p><b>Unerwarteter Fehler in<br/>"
               "%1</b></p>"
               "<p>Der Eintrag „%2“ konnte nicht gefunden werden. Bestenfalls funktioniert "
               "alles weiter, aber womöglich müssen die Netzwerkverbindung erneut gestartet "
               "und die Daten neu synchronisiert werden.</p>"
               "<p><b>Das sollte nicht passieren!</b></p>"
               "<p>Bitte einen Bugreport auf "
               "<a href=\"%3\">%3</a> schreiben!</p>").arg(
               QString::fromUtf8(funcInfoText), text, Urls::bugtracker));
    });
}

void RegistrationPage::setDisqualified()
{
    int round = -1;
    const auto selectedName = m_listView->selectedName();

    if (m_listModel->data(m_listView->selectedRow()).disqualified == 0) {
        auto dialog = DisqualificationDialog(this, selectedName,
                                                m_tournamentSettings->isPairMode(),
                                                m_db->allRounds());
        if (! dialog.exec()) {
            return;
        }
        round = dialog.round();

    } else {
        if (QMessageBox::question(this,
                tr("Disqualifikation zurücknehmen"),
                tr("Soll die Disqualifikation für %1 „%2“ wirklich zurückgenommen werden?").arg(
                m_tournamentSettings->isPairMode() ? tr("das Paar") : tr("den Spieler"),
                selectedName),
                QMessageBox::Yes | QMessageBox::Cancel,
                QMessageBox::Cancel) == QMessageBox::Cancel) {

            return;
        }
        round = 0;
    }

    Q_ASSERT(round != -1);

    // Someone could have deleted or renamed or edited entry
    if ((m_isClient || m_isServer ) && ! m_listModel->hasExactMatch(selectedName)) {
        QMessageBox::warning(this,
            tr("Kollidierende Netzwerkänderung"),
            tr("%1 „%2“ wurde mittlerweile von einem anderen Netzwerkteilnehmer entweder "
                "umbenannt oder gelöscht. Die Disqualifikation ist nicht mehr möglich und wird "
                "abgebrochen.").arg(
                m_tournamentSettings->isPairMode() ? tr("Das Paar") : tr("Der Spieler"),
                selectedName));
        return;
    }

    if (m_isClient) {
        m_client->requestSetDisqualified(selectedName, round);
        return;
    }

    QApplication::setOverrideCursor(Qt::WaitCursor);

    if (round != 0) {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Disqualifiziere Paar „%1“ ab Runde %2 …")
                             : tr("Disqualifiziere Spieler „%1“ ab Runde %2 …")).arg(
                                  selectedName, QString::number(round)));
    } else {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Nehme die Disqualifikation von Paar „%1“ zurück …")
                             : tr("Nehme die Disqualifikation von Spieler „%1“ zurück …")).arg(
                                  selectedName));
    }

    if (! m_db->setDisqualified(selectedName, round)) {
        return;
    }

    updatePlayersList();

    if (round != 0) {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Paar „%1“ ab Runde %2 disqualifiziert")
                             : tr("Spieler „%1“ ab Runde %2 disqualifiziert")).arg(
                                  selectedName, QString::number(round)));
    } else {
        Q_EMIT statusUpdate((m_tournamentSettings->isPairMode()
                             ? tr("Disqualifikation von Paar „%1“ zurückgenommen")
                             : tr("Disqualifikation von Spieler „%1“ zurückgenommen")).arg(
                                  selectedName));
    }

    Q_EMIT disqualificationChanged();
    QApplication::restoreOverrideCursor();

    if (m_isServer) {
        m_server->broadcastSetDisqualified(selectedName, round);
    }
}

void RegistrationPage::requestSetColumnVisible(rc::Column column, bool state)
{
    m_showColumn.value(column)->setChecked(state);
}

void RegistrationPage::requestSetHeaderVisible(bool state)
{
    m_hideHeaderAction->setChecked(state);
}

RegistrationListModel *RegistrationPage::listModel() const
{
    return m_listModel;
}

const QVector<Players::Data> &RegistrationPage::playersData() const
{
    return m_listModel->playersData();
}

void RegistrationPage::searchAsYouTypeToggled(bool state)
{
    m_settings->saveSearchAsYouTypeState(state);
    m_playersName->setClearButtonEnabled(state);
    m_searchWidget->setVisible(! state);
    if (! state) {
        m_searchWidget->clear();
    } else {
        m_searchWidget->setSearchText(m_playersName->text());
    }
}

void RegistrationPage::playersNameChanged(const QString &text)
{
    if (m_searchAsYouType->isChecked()) {
        m_searchWidget->setSearchText(text);
    }
}

void RegistrationPage::registerPhoneticSearchToggled()
{
    if (m_searchAsYouType->isChecked()) {
        updateSearch(m_playersName->text());
    }

    if (m_registerPhoneticSearch->isChecked()) {
        m_registerPhoneticSearch->setStyleSheet(QStringLiteral("QToolButton { padding: 2px; }"));
        m_registerPhoneticSearch->setToolTip(
            tr("<p>Phonetische Suche für die Suche nach Duplikaten bei der Anmeldung "
            "(de)aktivieren</p>")
            + PhoneticSearchButton::description());
    } else {
        m_registerPhoneticSearch->setStyleSheet(QStringLiteral(
            "QToolButton { padding: 2px; background: rgba(255, 0, 0, 50) };"));
        m_registerPhoneticSearch->setToolTip(tr(
            "<p>Phonetische Suche für die Suche nach Duplikaten bei der Anmeldung (de)aktivieren"
            "</p>"
            "<p><b>Vorsicht: Die phonetische Suche ist deaktiviert!</b></p>"
            "<p>Identische Anmeldungen mit kleinen Tippfehlern bzw. anderer Schreibweise des "
            "Namens werden womöglich nicht als Duplikat gemeldet!</p>"
            "<p>Z. B. würde die phonetische Suche „Schmidt Yannick / Meier Stefan“ als "
            "potentielles Duplikat von „Stephan Meyer / Jannik Schmitt“ finden. Ohne phonetische "
            "Suche wird die Eingabe ohne Warnung als neues Paar akzeptiert.</p>"));
    }
}
