// SPDX-FileCopyrightText: 2023 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes

#include "ScheduleDisplay.h"
#include "ResizableLabel.h"
#include "MarginSpinBox.h"

#include "SharedObjects/Settings.h"

#include "shared/Json.h"
#include "shared/Logging.h"
#include "shared/ColorHelper.h"

// Qt includes
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QDebug>
#include <QPushButton>
#include <QMenu>
#include <QLabel>
#include <QWidgetAction>
#include <QPalette>
#include <QColorDialog>
#include <QJsonObject>
#include <QJsonArray>
#include <QMessageBox>
#include <QWindow>
#include <QScreen>

static constexpr int s_settingsVersion = 1;

static const QLatin1String s_colors("colors");
static const QLatin1String s_background("background");
static const QLatin1String s_text("text");
static const QLatin1String s_negativeTime("negative time");
static const QLatin1String s_margins("margins");
static const QLatin1String s_labelSettings("label settings");
static const QLatin1String s_alignment("alignment");
static const QLatin1String s_bold("bold");
static const QLatin1String s_maxHeight("max height");
static const QLatin1String s_stretch("stretch");

static const QHash<Qt::AlignmentFlag, QString> s_alignmentToString {
    { Qt::AlignLeft,   QLatin1String("left")   },
    { Qt::AlignCenter, QLatin1String("center") },
    { Qt::AlignRight,  QLatin1String("right")  }
};

static const QHash<QString, Qt::AlignmentFlag> s_alignmentFromString {
    { QLatin1String("left"),   Qt::AlignLeft   },
    { QLatin1String("center"), Qt::AlignCenter },
    { QLatin1String("right"),  Qt::AlignRight  }
};

static const QLatin1String s_clock("clock");
static const QLatin1String s_roundOrBreak("round or break");
static const QLatin1String s_roundOrBreakTimeLeft("round or break time left");
static const QLatin1String s_nextRoundStart("next round start");

ScheduleDisplay::ScheduleDisplay(QWidget *parent, Settings *settings)
    : QDialog(parent),
      m_settings(settings)
{
    setAttribute(Qt::WA_DeleteOnClose, true);
    setWindowTitle(tr("Zeitplan-Info-Display"));

    auto *wrapperLayout = new QHBoxLayout(this);
    wrapperLayout->setContentsMargins(0, 0, 0, 0);
    const auto originalSpacing = wrapperLayout->spacing();
    wrapperLayout->setSpacing(0);

    auto *menuLayout = new QVBoxLayout;
    menuLayout->setSpacing(originalSpacing);
    wrapperLayout->addLayout(menuLayout);

    auto *menuButton = new QPushButton(tr("…"));
    menuButton->setMaximumWidth(50);
    menuButton->setFlat(true);
    menuButton->setFocusPolicy(Qt::NoFocus);
    menuButton->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum);

    auto *menu = new QMenu(this);

    m_toggleFullScreen = menu->addAction(tr("Vollbildanzeige"));
    m_toggleFullScreen->setCheckable(true);
    connect(m_toggleFullScreen, &QAction::triggered, this, &ScheduleDisplay::toggleFullScreen);

    menu->addSeparator();

    auto *colorsMenu = menu->addMenu(tr("Farben"));

    auto *baseAction = colorsMenu->addAction(tr("Hintergrund"));
    connect(baseAction, &QAction::triggered,
            std::bind(&ScheduleDisplay::chooseColor, this, s_background));

    auto *windowTextAction = colorsMenu->addAction(tr("Text"));
    connect(windowTextAction, &QAction::triggered,
            std::bind(&ScheduleDisplay::chooseColor, this, s_text));

    auto *brightTextAction = colorsMenu->addAction(tr("Negative Zeit"));
    connect(brightTextAction, &QAction::triggered,
            std::bind(&ScheduleDisplay::chooseColor, this, s_negativeTime));

    m_setFormattingVisible = menu->addAction(tr("Formatierungsoptionen anzeigen"));
    m_setFormattingVisible->setCheckable(true);
    connect(m_setFormattingVisible, &QAction::toggled, this,
            &ScheduleDisplay::setFormattingVisible);

    auto *marginsMenu = menu->addMenu(tr("Ränder"));

    for (auto pos : { Qt::AlignTop, Qt::AlignRight, Qt::AlignBottom, Qt::AlignLeft }) {
        auto *action = new QWidgetAction(this);
        auto *spinBox = new MarginSpinBox(pos, this);
        connect(spinBox, &MarginSpinBox::valueChanged, this, &ScheduleDisplay::setMargin);
        m_margins.insert(pos, spinBox);
        action->setDefaultWidget(spinBox);
        marginsMenu->addAction(action);
    }

    menu->addSeparator();

    auto *settingsMenu = menu->addMenu(tr("Einstellungen"));

    auto *saveSettingsAction = settingsMenu->addAction(tr("Speichern"));
    connect(saveSettingsAction, &QAction::triggered, this, &ScheduleDisplay::saveSettings);

    auto *resetSettingsAction = settingsMenu->addAction(tr("Zurücksetzen"));
    connect(resetSettingsAction, &QAction::triggered, this, &ScheduleDisplay::resetSettings);

    menu->addSeparator();

    auto *closeAction = menu->addAction(tr("Fenster schließen"));
    connect(closeAction, &QAction::triggered, this, &QDialog::accept);

    menuButton->setMenu(menu);

    menuLayout->addWidget(menuButton);
    menuLayout->addStretch();

    auto *leftSpacerLayout = new QVBoxLayout;
    wrapperLayout->addLayout(leftSpacerLayout);
    auto *leftSpacer = new QLabel;
    leftSpacer->setMaximumWidth(0);
    leftSpacer->setMinimumWidth(0);
    leftSpacer->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Ignored);
    m_spacers.insert(Qt::AlignLeft, leftSpacer);
    leftSpacerLayout->addWidget(leftSpacer);

    m_labelsWrapper = new QWidget;
    auto *labelsLayout = new QVBoxLayout(m_labelsWrapper);
    labelsLayout->setContentsMargins(0, 0, 0, 0);
    labelsLayout->setSpacing(originalSpacing);
    wrapperLayout->addWidget(m_labelsWrapper);

    auto *topSpacer = new QLabel;
    topSpacer->setMaximumHeight(0);
    topSpacer->setMinimumHeight(0);
    m_spacers.insert(Qt::AlignTop, topSpacer);
    topSpacer->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    labelsLayout->addWidget(topSpacer);

    auto *clock = new ResizableLabel;
    m_labels.insert(s_clock, clock);
    connect(clock, &ResizableLabel::stretchFactorChanged,
            std::bind(QOverload<QWidget *, int>::of(&QVBoxLayout::setStretchFactor),
                      labelsLayout, clock, std::placeholders::_1));
    labelsLayout->addWidget(clock);

    auto *roundOrBreak = new ResizableLabel;
    m_labels.insert(s_roundOrBreak, roundOrBreak);
    connect(roundOrBreak, &ResizableLabel::stretchFactorChanged,
            std::bind(QOverload<QWidget *, int>::of(&QVBoxLayout::setStretchFactor),
                      labelsLayout, roundOrBreak, std::placeholders::_1));
    labelsLayout->addWidget(roundOrBreak);

    auto *roundOrBreakTimeLeft = new ResizableLabel;
    m_labels.insert(s_roundOrBreakTimeLeft, roundOrBreakTimeLeft);
    connect(roundOrBreakTimeLeft, &ResizableLabel::stretchFactorChanged,
            std::bind(QOverload<QWidget *, int>::of(&QVBoxLayout::setStretchFactor),
                      labelsLayout, roundOrBreakTimeLeft, std::placeholders::_1));
    labelsLayout->addWidget(roundOrBreakTimeLeft);

    auto *nextRoundStart = new ResizableLabel;
    m_labels.insert(s_nextRoundStart, nextRoundStart);
    connect(nextRoundStart, &ResizableLabel::stretchFactorChanged,
            std::bind(QOverload<QWidget *, int>::of(&QVBoxLayout::setStretchFactor),
                      labelsLayout, nextRoundStart, std::placeholders::_1));
    labelsLayout->addWidget(nextRoundStart);

    auto *bottomSpacer = new QLabel;
    bottomSpacer->setMaximumHeight(0);
    bottomSpacer->setMinimumHeight(0);
    bottomSpacer->setSizePolicy(QSizePolicy::Ignored, QSizePolicy::Maximum);
    m_spacers.insert(Qt::AlignBottom, bottomSpacer);
    labelsLayout->addWidget(bottomSpacer);

    auto *rightSpacerLayout = new QVBoxLayout;
    wrapperLayout->addLayout(rightSpacerLayout);
    auto *rightSpacer = new QLabel;
    rightSpacer->setSizePolicy(QSizePolicy::Maximum, QSizePolicy::Ignored);
    m_spacers.insert(Qt::AlignRight, rightSpacer);
    rightSpacerLayout->addWidget(rightSpacer);

    loadSettings();
}

void ScheduleDisplay::updateClock(const QString &text)
{
    m_labels[s_clock]->setText(text);
}

void ScheduleDisplay::updateCurrentRound(int round)
{
    m_currentRound = round;

    if (! m_scheduleStarted) {
        m_labels[s_roundOrBreak]->setText(tr("Keine laufende Runde"));
    } else if (m_currentRound == 0) {
        m_labels[s_roundOrBreak]->setText(tr("Pause"));
    } else {
        m_labels[s_roundOrBreak]->setText(tr("Runde %1").arg(m_currentRound));
    }
}

void ScheduleDisplay::updateNextRound(int round)
{
    m_nextRound = round;
}

void ScheduleDisplay::setTimeToNextRoundText(const QString &text)
{
    m_labels[s_roundOrBreakTimeLeft]->setText(
        tr("Verbleibend: %1").arg(text.isEmpty() ? tr("–") : text));

    const auto isNegative = text.startsWith(tr("- "));
    if (isNegative && ! m_negativeTime) {
        m_labels[s_roundOrBreakTimeLeft]->setStyleSheet(
            QStringLiteral("color: %1").arg(m_colors[s_negativeTime].name()));
        m_negativeTime = true;
    } else if (! isNegative && m_negativeTime) {
        m_labels[s_roundOrBreakTimeLeft]->setStyleSheet(
            QStringLiteral("color: %1").arg(m_colors[s_text].name()));
        m_negativeTime = false;
    }
}

void ScheduleDisplay::updateRoundTimeLeft(const QString &text)
{
    if (m_currentRound != 0) {
        setTimeToNextRoundText(text);
    }
}

void ScheduleDisplay::updateTimeToNextRound(const QString &text)
{
    if (m_currentRound == 0) {
        setTimeToNextRoundText(text);
    }
}

void ScheduleDisplay::updateNextRoundStart(const QString &text)
{
    m_labels[s_nextRoundStart]->setText(
        tr("Beginn der nächsten Runde: %1").arg(text.isEmpty() ? tr("–") : text));
}

void ScheduleDisplay::toggleFullScreen()
{
    const auto currentWindowState = windowState();
    if (currentWindowState != Qt::WindowFullScreen) {
        m_lastWindowState = currentWindowState;
        setWindowState(Qt::WindowFullScreen);
    } else {
        setWindowState(m_lastWindowState);
    }
}

void ScheduleDisplay::setFormattingVisible(bool state)
{
    for (auto *label : m_labels.values()) {
        label->setFormattingVisible(state);
    }

    for (auto pos : { Qt::AlignTop, Qt::AlignRight, Qt::AlignBottom, Qt::AlignLeft }) {
        m_spacers[pos]->setStyleSheet(state ? QStringLiteral("background-color: red;")
                                            : QStringLiteral("background-color: transparent;"));
    }
}

void ScheduleDisplay::setMargin(Qt::Alignment pos, int margin)
{
    auto *spacer = m_spacers[pos];
    if (pos == Qt::AlignLeft || pos == Qt::AlignRight) {
        spacer->setMinimumWidth(margin);
        spacer->setMaximumWidth(margin);
    } else {
        spacer->setMinimumHeight(margin);
        spacer->setMaximumHeight(margin);
    }
}

void ScheduleDisplay::chooseColor(const QString &name)
{
    Q_ASSERT(name == s_background || name == s_text || name == s_negativeTime);

    const auto color = QColorDialog::getColor(m_colors.value(name));

    if (color.isValid()) {
        m_colors[name] = color;
        updateColors();
    }
}

void ScheduleDisplay::updateColors()
{
    auto palette = QPalette();
    palette.setColor(QPalette::Window, m_colors.value(s_background));
    setAutoFillBackground(true);
    setPalette(palette);
    show();

    m_labelsWrapper->setStyleSheet(QStringLiteral("QLabel { color: %1; }").arg(
                                                  m_colors.value(s_text).name()));

    setFormattingVisible(m_setFormattingVisible->isChecked());
}

void ScheduleDisplay::saveSettings()
{
    QJsonObject settings;

    // Colors
    QJsonObject colorsObject;
    for (const auto &colorName : { s_background, s_text, s_negativeTime }) {
        colorsObject.insert(colorName, ColorHelper::colorToString(m_colors.value(colorName)));
    };
    settings.insert(s_colors, colorsObject);

    // Margins

    // Just like with CSS's shorthand margin: top right bottom left
    settings.insert(s_margins, QJsonArray { m_spacers.value(Qt::AlignTop)->maximumHeight(),
                                            m_spacers.value(Qt::AlignRight)->maximumWidth(),
                                            m_spacers.value(Qt::AlignBottom)->maximumHeight(),
                                            m_spacers.value(Qt::AlignLeft)->maximumWidth() });

    // Label settings

    QJsonObject labelSettings;

    for (const auto &id : { s_clock, s_roundOrBreak, s_roundOrBreakTimeLeft, s_nextRoundStart }) {
        QJsonObject currentSettings;
        const auto *label = m_labels.value(id);
        currentSettings.insert(s_alignment, s_alignmentToString.value(label->alignment()));
        currentSettings.insert(s_bold,      label->isBold());
        currentSettings.insert(s_maxHeight, label->maxHeight());
        currentSettings.insert(s_stretch,   label->stretch());
        labelSettings.insert(id, currentSettings);
    }

    settings.insert(s_labelSettings, labelSettings);

    m_settings->saveScheduleDisplaySettings(Json::serialize(settings), s_settingsVersion);

    QMessageBox::information(this, tr("Einstellungen speichern"), tr("Einstellungen gespeichert!"));
}

void ScheduleDisplay::resetSettings()
{
    m_colors[s_background]   = Qt::white;
    m_colors[s_text]         = Qt::black;
    m_colors[s_negativeTime] = Qt::red;
    updateColors();

    m_margins[Qt::AlignTop]->setValue(20);
    m_margins[Qt::AlignRight]->setValue(50);
    m_margins[Qt::AlignBottom]->setValue(20);
    m_margins[Qt::AlignLeft]->setValue(0);

    ResizableLabel *label;

    label = m_labels.value(s_clock);
    label->setAlignment(Qt::AlignRight);
    label->setBold(false);
    label->setMaxHeight(70);
    label->setStretch(1);

    label = m_labels.value(s_roundOrBreak);
    label->setAlignment(Qt::AlignCenter);
    label->setBold(true);
    label->setMaxHeight(230);
    label->setStretch(6);

    label = m_labels.value(s_roundOrBreakTimeLeft);
    label->setAlignment(Qt::AlignCenter);
    label->setBold(false);
    label->setMaxHeight(190);
    label->setStretch(5);

    label = m_labels.value(s_nextRoundStart);
    label->setAlignment(Qt::AlignCenter);
    label->setBold(false);
    label->setMaxHeight(120);
    label->setStretch(5);
}

void ScheduleDisplay::loadSettings()
{
    const auto version = m_settings->scheduleDisplaySettingsVersion();
    const auto settings = m_settings->scheduleDisplaySettings();

    if (settings.isEmpty() && version == 0) {
        qCDebug(MuckturnierLog) << "No schedule display settings saved yet"
                                << "- falling back to defaults";
        resetSettings();
        return;
    }

    if (version != s_settingsVersion) {
        qCDebug(MuckturnierLog) << "Schedule display settings version is" << version
                                << "expected:" << s_settingsVersion
                                << "- falling back to defaults";
        resetSettings();
        return;
    }

    if (settings.isEmpty()) {
        qCDebug(MuckturnierLog) << "No schedule display settings found"
                                << "- falling back to defaults";
        resetSettings();
        return;
    }

    const auto parsed = Json::objectFromString(settings);

    // Colors

    const auto colors = parsed.value(s_colors).toObject();

    for (const auto &colorName : { s_background, s_text, s_negativeTime }) {
        const auto color = colors.value(colorName);
        if (color.isUndefined()) {
            qCDebug(MuckturnierLog) << "color for" << colorName
                                    << "is not defined - falling back to defaults";
            resetSettings();
            return;
        }
        m_colors[colorName] = ColorHelper::colorFromString(color.toString());
    }

    updateColors();

    // Margins

    const auto marginsValue = parsed.value(s_margins);
    if (marginsValue.isUndefined()) {
        qCDebug(MuckturnierLog) << "Margins not defined - falling back to defaults";
        resetSettings();
        return;
    }

    const auto margins = marginsValue.toArray();
    if (margins.count() != 4) {
        qCDebug(MuckturnierLog) << "Margins definition invalid - falling back to defaults";
        resetSettings();
        return;
    }

    m_margins[Qt::AlignTop]->setValue(margins.at(0).toInt());
    m_margins[Qt::AlignRight]->setValue(margins.at(1).toInt());
    m_margins[Qt::AlignBottom]->setValue(margins.at(2).toInt());
    m_margins[Qt::AlignLeft]->setValue(margins.at(3).toInt());

    // Labels

    const auto labelSettingsValue = parsed.value(s_labelSettings);
    if (labelSettingsValue.isUndefined()) {
        qCDebug(MuckturnierLog) << "Label settings not found - falling back to defaults";
        resetSettings();
        return;
    }

    const auto labelSettings = labelSettingsValue.toObject();

    for (const auto id : { s_clock, s_roundOrBreak, s_roundOrBreakTimeLeft, s_nextRoundStart} ) {
        const auto settingsValue = labelSettings.value(id);
        if (settingsValue.isUndefined()) {
            qCDebug(MuckturnierLog) << "Label settings for" << id
                                    << "not found - falling back to defaults";
            resetSettings();
            return;
        }

        const auto settings = settingsValue.toObject();
        auto *label = m_labels.value(id);

        // Alignment

        const auto alignmentValue = settings.value(s_alignment);
        if (alignmentValue.isUndefined()) {
            qCDebug(MuckturnierLog) << "No alignment setting found for" << id
                                    << "- falling back to defaults";
            resetSettings();
            return;
        }

        const auto alignment = alignmentValue.toString();
        if (! s_alignmentFromString.contains(alignment)) {
            qCDebug(MuckturnierLog) << "Alignment setting" << alignment << "for" << id
                                    << "is invalid - falling back to defaults";
            resetSettings();
            return;
        }
        label->setAlignment(s_alignmentFromString.value(alignment));

        // Bold
        const auto boldValue = settings.value(s_bold);
        if (boldValue.isUndefined()) {
            qCDebug(MuckturnierLog) << "No bold setting found for" << id
                                    << "- falling back to defaults";
            resetSettings();
            return;
        }
        label->setBold(boldValue.toBool());

        // Max height
        const auto maxHeightValue = settings.value(s_maxHeight);
        if (maxHeightValue.isUndefined()) {
            qCDebug(MuckturnierLog) << "No max height setting found for" << id
                                    << "- falling back to defaults";
            resetSettings();
            return;
        }
        label->setMaxHeight(maxHeightValue.toInt());

        // Stretch
        const auto stretchValue = settings.value(s_stretch);
        if (stretchValue.isUndefined()) {
            qCDebug(MuckturnierLog) << "No stretch setting found for" << id
                                    << "- falling back to defaults";
            resetSettings();
            return;
        }
        label->setStretch(stretchValue.toInt());
    }

    qCDebug(MuckturnierLog) << "Successfully loaded the saved display settings";
}

void ScheduleDisplay::showFullScreen(QScreen *screen)
{
    setWindowState(Qt::WindowNoState);
    windowHandle()->setScreen(screen);
    move(screen->geometry().topLeft());
    m_toggleFullScreen->setChecked(true);
    toggleFullScreen();
}

void ScheduleDisplay::setScheduleStarted(bool state)
{
    m_scheduleStarted = state;
}
