// SPDX-FileCopyrightText: 2010-2023 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef DATABASE_H
#define DATABASE_H

// Local includes

#include "shared/StructTemplate.h"
#include "shared/Markers.h"
#include "shared/Draw.h"
#include "shared/Scores.h"
#include "shared/Players.h"

// Qt includes
#include <QObject>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <QMap>
#include <QColor>
#include <QHash>

// Local classes
class SharedObjects;
class StringTools;
class TournamentSettings;

// Qt classes
class QLockFile;

using namespace StructTemplate;

class Database : public QObject
{
    Q_OBJECT

public:
    enum Error {
        NoError,
        CouldNotOpen,
        DbVersion3,
        DbVersion4,
        DbVersion5,
        DbVersion6,
        DbVersion7,
        DbVersion8,
        DbVersion9,
        DbVersion10
    };

    struct Assignment
    {
        QHash<int, QVector<int>> idsForTable;
        QHash<int, int> tableForId;
    };

    struct BoogerResult
    {
        int pairOrPlayer;
        int score;
    };

    struct RankData
    {
        int rank;
        QString name;
        int boogers;
        int goals;
        int opponentGoals;
        int playedBoogers;
    };

    struct GoalsImportant
    {
        QVector<bool> goals;
        QVector<bool> opponentGoals;
    };

    explicit Database(QObject *parent, SharedObjects *sharedObjects, const QString *namesSeparator);

    // Underlying database handling
    QString checkForLockFile(const QString &dbFile) const;
    bool create(const QString &dbFile);
    bool saveSettings();
    Database::Error openDatabase(const QString &dbFile);
    Database::Error checkDatabase(const QString &dbFile);
    bool updateDatabase(const QString &dbFileName, Error dbVersion);
    void closeDatabaseConnection();
    bool reset();

    // Settings
    void setNetworkTournamentStarted(bool state);
    bool networkTournamentStarted() const;
    bool localTournamentStarted() const;
    void setPairsOrPlayersCount(int count);

    // Functions returning const values/references or boolean operations on them

    bool isWritable() const;
    bool isChangeable() const;
    bool tournamentStarted() const;
    int dbVersion() const;
    int necessaryDivisor() const;
    bool playersDividable() const;
    bool playersDividable(int count) const;
    int pairsOrPlayersCount() const;
    int allRounds() const;
    bool isOpen() const;
    bool allRoundsFinished() const;
    GoalsImportant goalsImportance(const QVector<Database::RankData> &ranking) const;

    const QString &errorMessage() const;
    const QString &dbFile() const;
    const QString &dbPath() const;
    const QMap<int, bool> &roundsState() const;

    // Functions using SQL to write to the database

    bool addMarkers(const QVector<Markers::Marker> &markers);
    bool registerPlayers(const QString &name, int markerId);
    bool registerPlayers(const QStringList &names);
    bool renamePlayers(const QString &oldName, const QString &newName, int newMarker);
    bool renamePlayers(const QStringList &oldNames, const QStringList &newNames);
    bool updateDraw(const QString &name, const Draw::Seat &seat, int marker);
    bool setListDraw(const QVector<QString> &names, const QVector<Draw::Seat> &seats);
    bool assemblePair(const QString &player1, const QString &player2, const QString &separator,
                      int assignedMarker);
    bool deletePlayers(const QString &name);
    bool deleteMarkedPlayers(int markerId);
    bool deleteUnmarkedPlayers();
    bool deleteAllPlayers();
    bool deleteAllDraws();

    bool updateMarker(const QString &name, int markerId);
    bool listUpdateMarker(int markerId);
    bool listRemoveMarker(int markerId);

    bool addMarker(const Markers::Marker &marker);
    bool editMarker(int id, const QString &name, const QColor &color,
                    Markers::MarkerSorting sorting);
    bool deleteMarker(int id, int sequence);
    bool moveMarker(int id, int sequence, int direction);

    bool saveScore(int round, int table, const QVector<Database::BoogerResult> &score);
    bool deleteScore(int round, int table);

    bool adoptAllData(const QVector<Players::Data> &playersData,
                      const QVector<Markers::Marker> &markers);
    bool adoptPlayersMetadata(const QVector<Players::Data> &playersData,
                              const QVector<Markers::Marker> &markers);

    bool setDisqualified(const QString &name, int round);

    bool saveScheduleTimes(int round, const QString &start, const QString &end);
    bool resetScheduleTimes(int fromRound);

    // Functions using SQL to read from the database
    TSuccess<QStringList> players();
    TSuccess<QVector<Players::Data>> playersData();
    TSuccess<QVector<int>> missingTables(int round);
    TSuccess<QHash<QString, int>> missingPlayers(int round);
    TSuccess<Assignment> assignment(int round);
    TSuccess<Scores::Score> scores(int round);
    TSuccess<QVector<BoogerResult>> singleScore(int round, int table);
    TSuccess<QVector<RankData>> ranking(int upToRound);
    TSuccess<QVector<Markers::Marker>> markers();
    TSuccess<QMap<int, QVector<QString>>> disqualifiedPlayers(int round);
    TSuccess<QVector<QString>> getScheduleData();

Q_SIGNALS:
    void sqlError(const QString &text);
    void roundsStateChanged();

private: // Functions
    // SQL helper functions
    bool startTransaction();
    bool commitTransaction();
    bool rollbackTransaction(QSqlQuery &query);
    bool queryPrepare(QSqlQuery &query, const QString &text);
    bool queryExec(QSqlQuery &query);
    bool queryExec(QSqlQuery &query, const QString &text);
    bool queryExecBatch(QSqlQuery &query);
    QString sqlErrorText() const;
    QString sqlErrorText(const QSqlQuery &query) const;
    void emitSqlError(const QString &text);
    QVariantList stringListToVariantList(const QStringList &stringList) const;

    // Helper functions for catching SQL errors

    template<typename SqlFunction>
    bool writeHelper(SqlFunction sqlFunction)
    {
        if (! startTransaction()) {
            return false;
        }

        QSqlQuery query(m_db);
        try {
            sqlFunction(query);
            commitTransaction();
        } catch (...) {
            rollbackTransaction(query);
            return false;
        }

        return true;
    }

    template<typename SqlFunction>
    void readHelper(SqlFunction sqlFunction, bool &success)
    {
        try {
            QSqlQuery query(m_db);
            sqlFunction(query);
            success = true;
        } catch (...) {
            success = false;
        }
    }

    // Connecting helpers
    bool connectDatabase(const QString &dbFile);
    bool lockDatabase();
    QString lockFileName(const QString &dbFile) const;
    bool lockDatabase(const QString &dbFileName);
    void updateDbFileInfo();
    bool checkIfDbIsLocked();

    // Database revision update functions
    void updateDatabaseFromV3(QSqlQuery &query);
    void updateDatabaseFromV4(QSqlQuery &query);
    void updateDatabaseFromV5(QSqlQuery &query);
    void updateDatabaseFromV6(QSqlQuery &query);
    void updateDatabaseFromV7(QSqlQuery &query);
    void updateDatabaseFromV8(QSqlQuery &query);
    void updateDatabaseFromV9(QSqlQuery &query);
    void updateDatabaseFromV10(QSqlQuery &query);

    bool checkIfTournamentStarted();
    bool updateRoundsState();
    bool ranksDiffer(const RankData &currentRank, const RankData &lastRank);
    void updateAssignment(QSqlQuery &query,
                          int round, int pairOrPlayer, int pairNumber, int playerNumber, int table);
    void updateDraw(QSqlQuery &query, const QString &name, const Draw::Seat &seat, int marker);
    void addMarker(QSqlQuery &query, const Markers::Marker &marker);
    void updateMarker(QSqlQuery &query, const QString &name, int markerId);
    void registerPlayers(QSqlQuery &query, const QStringList &names);
    void listUpdateMarker(QSqlQuery &query, int markerId);
    void renamePlayers(QSqlQuery &query, const QString &oldName, const QString &newName,
                       int newMarker);
    void deletePlayers(QSqlQuery &query, const QString &name);
    void deleteAllDraws(QSqlQuery &query);
    void adoptPlayersMetadata(QSqlQuery &query,
                              const QVector<Players::Data> &playersData,
                              const QVector<Markers::Marker> &markers);
    void setDisqualified(QSqlQuery &query, const QString &name, int round);

private: // Variables
    TournamentSettings *m_tournamentSettings;
    StringTools *m_stringTools;
    const QString *m_namesSeparator;
    QColor m_foreground;

    QLockFile *m_lockFile = nullptr;
    QSqlDatabase m_db;
    QString m_errorMessage;
    QString m_dbFileName;
    QString m_dbFilePath;
    bool m_isWritable;
    bool m_emitSqlErrors = false;

    bool m_tournamentStarted = false;
    bool m_networkTournamentStarted = false;
    int m_pairsOrPlayersCount = 0;
    int m_scoresPerRound = 0;
    int m_tablesCount = 0;
    QMap<int, bool> m_roundsState;
    int m_allRounds = 0;

};

#endif // DATABASE_H
