// SPDX-FileCopyrightText: 2019-2023 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

#ifndef SETTINGS_H
#define SETTINGS_H

// Local includes

#include "shared/RegistrationColumns.h"
#include "shared/Tournament.h"
#include "shared/Markers.h"
#include "shared/TournamentTemplate.h"
#include "shared/Draw.h"
#include "shared/Schedule.h"

#include "SchedulePage/StopWatchState.h"

// Qt includes
#include <QString>
#include <QSettings>

// Qt classes
class QAction;

class Settings : public QSettings
{
    Q_OBJECT

public:
    static const QString registrationPhoneticId;

    explicit Settings(QObject *parent);

    void saveTabPosition(const QString &position);
    QString tabPosition() const;

    bool saveMainWindowGeometryEnabled() const;
    void enableSaveMainWindowGeometry(bool state);
    void saveMainWindowGeometry(const QByteArray &data);
    QByteArray mainWindowGeometry() const;

    bool saveMainWindowStateEnabled() const;
    void enableSaveMainWindowState(bool state);
    void saveMainWindowState(const QByteArray &data);
    QByteArray mainWindowState() const;

    void enableSaveLastDbPath(bool state);
    bool saveLastDbPathEnabled() const;
    void saveLastDbPath(const QString &dbPath);
    QString lastDbPath() const;

    void enableSaveStopWatchState(bool state);
    bool saveStopWatchStateEnabled() const;
    void saveStopWatchState(StopWatchState::State state);
    StopWatchState::State stopWatchState() const;

    bool newTournamentShowExtendedOptions() const;
    void saveNewTournamentShowExtendedOptions(bool state);

    bool scheduleStartStopWatch() const;
    void saveScheduleStartStopWatch(bool state);

    void saveScheduleDisplaySettings(const QString &settings, int version);
    int scheduleDisplaySettingsVersion() const;
    QString scheduleDisplaySettings() const;

    void enableSavePhoneticSearchState(bool state);
    bool savePhoneticSearchStateEnabled() const;
    bool phoneticSearchState(const QString &searchId) const;
    void savePhoneticSearchState(const QString &searchId, bool state);

    void enableSaveRegistrationOptions(bool state);
    bool saveRegistrationOptionsEnabled() const;
    bool searchAsYouTypeState() const;
    void saveSearchAsYouTypeState(bool state);
    bool autoCapitalizeState() const;
    void saveAutoCapitalizeState(bool state);

    void saveTournamentTemplate(const TournamentTemplate::Settings &tournamentTemplate);
    TournamentTemplate::Settings tournamentTemplate() const;

    void saveDbFileNameTemplate(const QString &text);
    QString dbFileNameTemplate() const;

    void saveMarkersTemplate(Tournament::Mode tournamentMode,
                             const QVector<Markers::Marker> &markers);
    QVector<Markers::Marker> defaultMarkersTemplate(Tournament::Mode tournamentMode) const;
    QVector<Markers::Marker> markersTemplate(Tournament::Mode tournamentMode) const;
    QHash<Markers::Type, int> specialMarkers() const;

    void saveMarker(Markers::Type type, int id);
    void saveMarkers(const QHash<Markers::Type, int> &markers);
    int marker(Markers::Type type) const;

    void saveSinglesManagementTemplate(const Markers::SinglesManagement &data);
    Markers::SinglesManagement defaultSinglesManagementTemplate() const;
    Markers::SinglesManagement singlesManagementTemplate() const;

    void saveBoogerSymbol(const QString &text);
    const QString &boogerSymbol() const;

    void saveNamesSeparator(const QString &text);
    void setTransientNamesSeparator(const QString &text);
    bool resetNamesSeparator();
    const QString &namesSeparator() const;

    void saveServerPort(int port);
    int serverPort() const;
    void saveConnectTimeout(int timeout);
    int connectTimeout() const;
    void saveRequestTimeout(int timeout);
    int requestTimeout() const;
    void saveDiscoverPort(int port);

    int discoverPort() const;
    void saveDiscoverTimeout(int timeout);
    int discoverTimeout() const;
    void saveDiscoverAttempts(int attempts);
    int discoverAttempts() const;

    void savePreferIpv6(bool state);
    bool preferIpv6() const;

    void saveRegistrationColumns(const QHash<RegistrationColumns::Column, bool> &columns);
    QHash<RegistrationColumns::Column, bool> registrationColumns() const;

    void saveDrawSettingsTemplate(const Draw::Settings &data);
    Draw::Settings drawSettingsTemplate() const;

    void saveScheduleSettingsTemplate(const Schedule::Settings &data);
    Schedule::Settings scheduleSettingsTemplate() const;

    void saveMessageDismissed(const QString &id, bool state);
    void showAllDismissedMessages();
    bool messageDismissed(const QString &id);

private: // Functions
    void saveSaveOnCloseValue(const QString &key, const QString &dataSet, bool state);
    void saveSaveOnCloseValue(const QString &key, const QVector<QString> &dataSets, bool state);
    bool saveOnCloseValue(const QString &key) const;
    bool parseMarkerData(int &id, const QJsonValue &value, QVector<Markers::Marker> &markers) const;

private: // Variables
    QString m_boogerSymbol;
    QString m_namesSeparator;
    QString m_originalNamesSeparator;

};

#endif // SETTINGS_H
