// SPDX-FileCopyrightText: 2010-2023 Tobias Leupold <tl at stonemx dot de>
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Local includes
#include "TableDelegate.h"

// Qt includes
#include <QPainter>
#include <QBrush>
#include <QDebug>
#include <QGuiApplication>
#include <QPalette>

TableDelegate::TableDelegate(QObject *parent, RowColorMode rowColorMode)
    : QStyledItemDelegate(parent),
      m_rowColorMode(rowColorMode)
{
    const bool darkMode = qApp->property("DARK_MODE").toBool();
    const QPalette palette;

    if (! darkMode) {
        m_topLineColor = palette.color(QPalette::Mid);
    } else {
        m_topLineColor = palette.color(QPalette::Light);
    }

#ifndef Q_OS_MACOS
    m_alternateBackground = QBrush(palette.color(QPalette::AlternateBase));
#else
    // If we're on macOS and in dark mode, QPalette::AlternateBase doesn't look nice.
    // We use a color derived from the default base in this case:
    if (! darkMode) {
        m_alternateBackground = QBrush(palette.color(QPalette::AlternateBase));
    } else {
        m_alternateBackground = QBrush(palette.color(QPalette::Base).lighter(140));
    }
#endif
}

void TableDelegate::drawBackground(QPainter *painter, const QStyleOptionViewItem &option,
                                   const QModelIndex &index) const
{
    if (   (m_rowColorMode == RowColorMode::AlternatingRows
            && index.row() % 2 != 0)

        || (m_rowColorMode == RowColorMode::AlternatingNRows
            && index.row() / m_alternationStep % 2 != 0)

        || (m_rowColorMode == RowColorMode::CustomAlternating
            && index.data(Display::DrawBackground).isValid()) ) {

        painter->fillRect(option.rect, m_alternateBackground);
    }
}

void TableDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option,
                          const QModelIndex &index) const
{
    painter->save();

    drawBackground(painter, option, index);

    bool doDrawTopLine = false;
    const auto row = index.row();

    if (m_rowColorMode != RowColorMode::CustomAlternating
        && row != 0
        && row != m_selectedRow + 1
        && (m_rowColorMode == RowColorMode::AlternatingRows || row % m_alternationStep == 0)) {

        doDrawTopLine = true;

    } else if (m_rowColorMode == RowColorMode::CustomAlternating) {
        doDrawTopLine = index.data(Display::DrawTopLine).isValid();
    }

    if (doDrawTopLine) {
        drawTopLine(painter, option);
    }

    painter->restore();

    drawText(painter, option, index);
}

void TableDelegate::drawTopLine(QPainter *painter, const QStyleOptionViewItem &option) const
{
    painter->setPen(m_topLineColor);
    painter->drawLine(option.rect.topLeft(), option.rect.topRight());
}

void TableDelegate::drawText(QPainter *painter, const QStyleOptionViewItem &option,
                             const QModelIndex &index) const
{
    QStyledItemDelegate::paint(painter, option, index);
}

void TableDelegate::setAlternationStep(int rows)
{
    m_alternationStep = rows;
}

void TableDelegate::setSelectedRow(int row)
{
    m_selectedRow = row;
}

int TableDelegate::selectedRow() const
{
    return m_selectedRow;
}
